#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import argparse
import functools
import sys
from typing import cast

from drgndoc.format import Formatter
from drgndoc.namespace import Namespace, ResolvedNode
from drgndoc.parse import Class, DocumentedNode, Node, parse_paths
from drgndoc.util import dot_join

escapes = []
for c in range(256):
    if c == 0:
        e = r"\0"
    elif c == 7:
        e = r"\a"
    elif c == 8:
        e = r"\b"
    elif c == 9:
        e = r"\t"
    elif c == 10:
        e = r"\n"
    elif c == 11:
        e = r"\v"
    elif c == 12:
        e = r"\f"
    elif c == 13:
        e = r"\r"
    elif c == 34:
        e = r"\""
    elif c == 92:
        e = r"\\"
    elif 32 <= c <= 126:
        e = chr(c)
    else:
        e = f"\\x{c:02x}"
    escapes.append(e)


def escape_string(s: str) -> str:
    return "".join([escapes[c] for c in s.encode("utf-8")])


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="generate docstring definitions for a C extension from Python source code/stub files"
    )
    parser.add_argument(
        "--header", "-H", action="store_true", help="generate header file"
    )
    parser.add_argument(
        "-m",
        "--module",
        dest="modules",
        metavar="MODULE[:NAME]",
        action="append",
        help="generate docstrings for the given module instead of all modules "
        "(may be given multiple times); "
        "an alternate name to use for the generated variables may also be given",
    )
    parser.add_argument(
        "paths", metavar="PATH", nargs="+", help="module or package path"
    )
    args = parser.parse_args()

    modules = parse_paths(args.paths, functools.partial(print, file=sys.stderr))
    namespace = Namespace(modules)
    formatter = Formatter(namespace)
    output_file = sys.stdout

    if args.header:
        output_file.write("/* Generated by drgndoc.docstrings -H. */\n\n")
    else:
        output_file.write("/* Generated by drgndoc.docstrings. */\n\n")

    def aux(resolved: ResolvedNode[Node], name: str) -> None:
        node = resolved.node
        if node.has_docstring():
            var_name = name.replace(".", "_") + "_DOC"
            if args.header:
                output_file.write("extern ")
            output_file.write(f"const char {var_name}[]")
            if not args.header:
                output_file.write(" =")
                lines = formatter.format(
                    cast(ResolvedNode[DocumentedNode], resolved),
                    name.rpartition(".")[2],
                    rst=False,
                )
                if lines:
                    for i, line in enumerate(lines):
                        output_file.write(f'\n\t"{escape_string(line)}')
                        if i != len(lines) - 1:
                            output_file.write("\\n")
                        output_file.write('"')
                else:
                    output_file.write(' ""')
            output_file.write(";\n")
        for attr in resolved.attrs():
            if isinstance(node, Class) and attr.name == "__init__":
                continue
            aux(attr, dot_join(name, attr.name))

    for module in args.modules or namespace.modules.keys():
        module, _, name = module.partition(":")
        resolved = namespace.resolve_global_name(module)
        if isinstance(resolved, ResolvedNode):
            aux(resolved, name or module)
        else:
            sys.exit(f"name {module} not found")



# Copyright (c) 2021, Oracle and/or its affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

from drgn import SymbolBinding, SymbolKind
from tests.linux_kernel import LinuxKernelTestCase, skip_unless_have_test_kmod


class TestSymbol(LinuxKernelTestCase):
    def test_global_symbol(self):
        symbol = self.prog.symbol("jiffies")
        self.assertEqual(symbol.name, "jiffies")
        self.assertEqual(symbol.binding, SymbolBinding.GLOBAL)
        self.assertEqual(symbol.kind, SymbolKind.OBJECT)

    @skip_unless_have_test_kmod
    def test_module_function_symbol(self):
        symbol = self.prog.symbol("drgn_test_function")
        self.assertEqual(symbol.name, "drgn_test_function")
        self.assertEqual(symbol.binding, SymbolBinding.GLOBAL)
        self.assertEqual(symbol.kind, SymbolKind.FUNC)

        symbol = self.prog.symbol(symbol.address)
        self.assertEqual(symbol.name, "drgn_test_function")
        self.assertEqual(symbol.binding, SymbolBinding.GLOBAL)
        self.assertEqual(symbol.kind, SymbolKind.FUNC)

    @skip_unless_have_test_kmod
    def test_module_data_symbol(self):
        symbol = self.prog.symbol("drgn_test_data")
        self.assertEqual(symbol.name, "drgn_test_data")
        self.assertEqual(symbol.binding, SymbolBinding.GLOBAL)
        self.assertEqual(symbol.kind, SymbolKind.OBJECT)

        symbol = self.prog.symbol(symbol.address)
        self.assertEqual(symbol.name, "drgn_test_data")
        self.assertEqual(symbol.binding, SymbolBinding.GLOBAL)
        self.assertEqual(symbol.kind, SymbolKind.OBJECT)

    @skip_unless_have_test_kmod
    def test_module_rodata_symbol(self):
        symbol = self.prog.symbol("drgn_test_rodata")
        self.assertEqual(symbol.name, "drgn_test_rodata")
        self.assertEqual(symbol.binding, SymbolBinding.GLOBAL)
        self.assertEqual(symbol.kind, SymbolKind.OBJECT)

        symbol = self.prog.symbol(symbol.address)
        self.assertEqual(symbol.name, "drgn_test_rodata")
        self.assertEqual(symbol.binding, SymbolBinding.GLOBAL)
        self.assertEqual(symbol.kind, SymbolKind.OBJECT)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

from drgn import Program
from tests import TestCase
from tests.resources import get_resource


class TestLinuxUserspaceCoreDump(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.prog = Program()
        cls.prog.set_enabled_debug_info_finders([])
        cls.prog.set_core_dump(get_resource("crashme.core"))
        cls.prog.load_debug_info([get_resource("crashme"), get_resource("crashme.so")])
        cls.trace = cls.prog.crashed_thread().stack_trace()

    @classmethod
    def tearDownClass(cls):
        del cls.trace
        del cls.prog

    def test_stack_frame_name(self):
        self.assertEqual(self.trace[0].name, "c")
        self.assertEqual(self.trace[5].name, "0x7f6112ad8088")
        self.assertEqual(self.trace[7].name, "_start")
        self.assertEqual(self.trace[8].name, "???")

    def test_stack_frame_function_name(self):
        self.assertEqual(self.trace[0].function_name, "c")
        self.assertIsNone(self.trace[5].function_name)
        self.assertIsNone(self.trace[7].function_name)
        self.assertIsNone(self.trace[8].function_name)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

from ctypes import c_int, c_int64, c_size_t, c_void_p
import mmap
import unittest

from _drgn_util.platform import NORMALIZED_MACHINE_NAME
from drgn.helpers.linux.boot import pgtable_l5_enabled
from tests.linux_kernel import LinuxKernelTestCase, _c, iter_maps


def first_available_slot(size, min_addr):
    for map in iter_maps():
        if map.start >= min_addr + size:
            break
        elif map.end >= min_addr:
            min_addr = map.end
    return min_addr


def can_mmap_high_address():
    mmap_func = _c.mmap
    mmap_func.argtypes = [c_void_p, c_size_t, c_int, c_int, c_int, c_int64]
    mmap_func.restype = c_void_p
    munmap_func = _c.munmap
    munmap_func.argtypes = [c_void_p, c_size_t]
    hint_addr = first_available_slot(mmap.PAGESIZE, 1 << 48)

    ret = mmap_func(
        hint_addr,
        mmap.PAGESIZE,
        mmap.PROT_READ | mmap.PROT_WRITE,
        # Ideally we would use MAP_FIXED, but its value is not exposed by the
        # mmap module, and it varies by architecture. Having identified a free
        # slot in our memory mappings (and hopefully not changing them since
        # then), we can be reasonably confident that we should get the address
        # we hinted anyway.
        mmap.MAP_PRIVATE | mmap.MAP_ANONYMOUS,
        -1,
        0,
    )
    if ret != c_void_p(-1).value:
        munmap_func(ret, mmap.PAGESIZE)
    return ret == hint_addr


class TestBoot(LinuxKernelTestCase):
    @unittest.skipUnless(NORMALIZED_MACHINE_NAME == "x86_64", "machine is not x86_64")
    def test_pgtable_l5_enabled(self):
        self.assertEqual(pgtable_l5_enabled(self.prog), can_mmap_high_address())



# Copyright (c) 2025, Oracle and/or its affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later
import contextlib
import io
import struct

from drgn import (
    Architecture,
    FaultError,
    Platform,
    Program,
    Symbol,
    SymbolBinding,
    SymbolKind,
)
from drgn.commands._crash._rd import _print_memory
from tests import TestCase


def add_fake_memory_segment(prog, start, bytes, is_phys=False):

    def reader(address, count, offset, physical):
        if physical == is_phys:
            return bytes[offset : offset + count]
        raise FaultError("unable to read memory", address)

    prog.add_memory_segment(start, len(bytes), reader, physical=is_phys)


class TestPrintMemory(TestCase):

    @classmethod
    def setUpClass(cls):
        # The only reason to set a specific architecture is to have a known word
        # size, which does impact alignment. The rest is not
        # architecture-specific.
        cls.prog = Program(platform=Platform(Architecture.X86_64))
        mem_data = bytes(range(8)) + struct.pack("<Q", 0x12345678) + bytes(range(8, 64))
        add_fake_memory_segment(cls.prog, 0xFFFF0000, mem_data)
        add_fake_memory_segment(cls.prog, 0x10000, b"ABCD", is_phys=True)
        cls.prog.register_symbol_finder(
            "test",
            lambda prog, name, address, one: (
                [
                    Symbol(
                        "test_sym",
                        0x12345678,
                        0,
                        SymbolBinding.GLOBAL,
                        SymbolKind.OBJECT,
                    )
                ]
                if (name == "test_sym" or (name is None and address == 0x12345678))
                else []
            ),
        )
        cls.prog.set_enabled_symbol_finders(["test"])

    def run_print_memory(self, address: int, count: int, *, unit: int, **kwargs):
        output = io.StringIO()
        physical = kwargs.pop("physical", False)
        memory = self.prog.read(address, count * unit, physical=physical)
        with contextlib.redirect_stdout(output):
            _print_memory(self.prog, address, memory, unit=unit, **kwargs)
        return output.getvalue()

    def test_print_memory_basic(self):
        result = self.run_print_memory(0xFFFF0000, 4, unit=1)
        self.assertEqual(
            result,
            """\
        ffff0000:  00 01 02 03                                       ....
""",
        )

    def test_unit_2(self):
        result = self.run_print_memory(0xFFFF0000, 2, unit=2)
        self.assertEqual(
            result,
            """\
        ffff0000:  0100 0302                                 ....
""",
        )

    def test_physical(self):
        result = self.run_print_memory(0x10000, 1, physical=True, unit=4)
        self.assertEqual(
            result,
            """\
           10000:  44434241                              ABCD
""",
        )

    def test_no_ascii(self):
        result = self.run_print_memory(0xFFFF0000, 4, unit=1, show_ascii=False)
        self.assertEqual(
            result,
            """\
        ffff0000:  00 01 02 03
""",
        )

    def test_annotate_symbols(self):
        result = self.run_print_memory(
            0xFFFF0008, 2, unit=8, annotate="symbols", show_ascii=False
        )
        self.assertEqual(
            result,
            """\
        ffff0008:  test_sym+0       0f0e0d0c0b0a0908
""",
        )

    def test_format_decimal(self):
        result = self.run_print_memory(0xFFFF0000, 1, unit=4, format="d")
        self.assertEqual(
            result,
            """\
        ffff0000:     50462976                                       ....
""",
        )

    def test_endian_big(self):
        result = self.run_print_memory(0x10000, 1, physical=True, unit=4, endian="big")
        self.assertEqual(
            result,
            """\
           10000:  41424344                              ABCD
""",
        )

    def test_annotate_wrong_unit(self):
        with self.assertRaises(ValueError):
            self.run_print_memory(0xFFFF0000, 1, unit=4, annotate="symbols")

    def test_annotate_wrong_format(self):
        with self.assertRaises(ValueError):
            self.run_print_memory(0xFFFF0000, 1, unit=4, format="d", annotate="symbols")



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later
from functools import cmp_to_key

from tests import TestCase
from util import KernelVersion, verrevcmp


class TestUtil(TestCase):
    def assertVersionSort(self, sorted_list):
        self.assertEqual(sorted(sorted_list, key=cmp_to_key(verrevcmp)), sorted_list)

    def test_verrevcmp(self):
        self.assertVersionSort(
            ["0~", "0", "1", "1.0", "1.1~rc1", "1.1~rc2", "1.1", "1.2", "1.12"]
        )
        self.assertVersionSort(["a", "."])
        self.assertVersionSort(["", "1"])
        self.assertVersionSort(["~", "~1"])
        self.assertVersionSort(["~~", "~~a", "~", "", "a"])

    def test_kernel_version(self):
        self.assertLess(KernelVersion("1.0"), KernelVersion("2.0"))
        self.assertLess(KernelVersion("5.6.0-rc6"), KernelVersion("5.6.0-rc7"))
        self.assertLess(KernelVersion("5.6.0-rc7"), KernelVersion("5.6.0"))
        self.assertLess(
            KernelVersion("5.6.0-rc7-vmtest2"), KernelVersion("5.6.0-vmtest1")
        )
        self.assertLess(KernelVersion("5.6.0-vmtest1"), KernelVersion("5.6.0-vmtest2"))



# Copyright (c) ByteDance, Inc. and its affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

from pathlib import Path
import unittest

from drgn.helpers.linux.nodemask import for_each_node, for_each_online_node, node_state
from tests.linux_kernel import LinuxKernelTestCase, parse_range_list

NODE_PATH = Path("/sys/devices/system/node")


@unittest.skipUnless(NODE_PATH.exists(), "kernel does not support NUMA")
class TestNodeMask(LinuxKernelTestCase):
    @staticmethod
    def _parse_node_list(name):
        return parse_range_list((NODE_PATH / name).read_text())

    def _test_for_each_node(self, func, name):
        self.assertEqual(list(func(self.prog)), sorted(self._parse_node_list(name)))

    def test_for_each_node(self):
        self._test_for_each_node(for_each_node, "possible")

    def test_for_each_online_node(self):
        self._test_for_each_node(for_each_online_node, "online")

    def _test_node_state(self, state_name, file_name):
        possible = self._parse_node_list("possible")
        expected = self._parse_node_list(file_name)
        state = self.prog[state_name]
        for node in possible:
            self.assertEqual(node_state(node, state), node in expected)

    def test_node_state(self):
        self._test_node_state("N_NORMAL_MEMORY", "has_normal_memory")
        # N_GENERIC_INITIATOR was added in Linux kernel commit 894c26a1c274
        # ("ACPI: Support Generic Initiator only domains") (in v5.10). Most of
        # the time it is unset, so if it exists we can use it to test the unset
        # case.
        if (NODE_PATH / "has_generic_initiator").exists():
            self._test_node_state("N_GENERIC_INITIATOR", "has_generic_initiator")



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import os
from pathlib import Path
import re
import signal

from drgn import Object
from drgn.helpers.linux.pid import find_task
from drgn.helpers.linux.signal import (
    decode_sigset,
    sigaction_flags,
    signal_names,
    signal_numbers,
    sigpending_for_each,
    sigset_to_hex,
)
from tests.linux_kernel import LinuxKernelTestCase, fork_and_stop


class TestSignal(LinuxKernelTestCase):
    def test_sigpending_for_each(self):
        with fork_and_stop(
            signal.pthread_sigmask, signal.SIG_BLOCK, {signal.SIGUSR1, signal.SIGUSR2}
        ) as (pid, _):
            os.kill(pid, signal.SIGUSR1)
            os.kill(pid, signal.SIGUSR2)
            task = find_task(self.prog, pid)
            self.assertLessEqual(
                {signal.SIGUSR1, signal.SIGUSR2},
                {
                    sigqueue.info.si_signo.value_()
                    for sigqueue in sigpending_for_each(
                        task.signal.shared_pending.address_of_()
                    )
                },
            )

    def test_signal_names(self):
        for name, number in signal_names(self.prog).items():
            try:
                expected = getattr(signal, name)
            except AttributeError:
                continue
            if name == "SIGRTMIN":
                # glibc uses some real-time signals internally and adjusts
                # SIGRTMIN accordingly.
                self.assertLessEqual(number, expected)
            else:
                self.assertEqual(number, expected)

    def test_signal_numbers(self):
        number_to_names = signal_numbers(self.prog)
        for name, number in signal_names(self.prog).items():
            self.assertIn(name, number_to_names[number])

    def test_sigaction_flags(self):
        # We don't have a good way to validate the values from Python, so just
        # sanity check the dictionary.
        for flag, value in sigaction_flags(self.prog).items():
            self.assertRegex(flag, r"^SA_")
            self.assertIsInstance(value, int)

    def test_decode_sigset(self):
        with fork_and_stop(
            signal.pthread_sigmask, signal.SIG_BLOCK, {signal.SIGUSR1, signal.SIGUSR2}
        ) as (pid, _):
            os.kill(pid, signal.SIGUSR1)
            os.kill(pid, signal.SIGUSR2)
            task = find_task(self.prog, pid)
            sigset = task.signal.shared_pending.signal
            self.assertRegex(decode_sigset(sigset), r"^\{.*\bSIGUSR1,SIGUSR2\b.*\}$")
            self.assertRegex(
                decode_sigset(sigset.address_of_()), r"^\{.*\bSIGUSR1,SIGUSR2\b.*\}$"
            )

    def test_decode_sigset_value(self):
        self.assertEqual(
            decode_sigset(
                self.prog, (1 << (signal.SIGHUP - 1)) | (1 << (signal.SIGTERM - 1))
            ),
            "{SIGHUP,SIGTERM}",
        )

    def test_decode_sigset_integer_object(self):
        self.assertEqual(
            decode_sigset(
                self.prog,
                Object(
                    self.prog,
                    "unsigned long",
                    (1 << (signal.SIGHUP - 1)) | (1 << (signal.SIGTERM - 1)),
                ),
            ),
            "{SIGHUP,SIGTERM}",
        )

    def test_sigset_to_hex(self):
        with fork_and_stop(
            signal.pthread_sigmask, signal.SIG_BLOCK, {signal.SIGUSR1, signal.SIGUSR2}
        ) as (pid, _):
            os.kill(pid, signal.SIGUSR1)
            os.kill(pid, signal.SIGUSR2)
            task = find_task(self.prog, pid)
            expected = re.search(
                r"^ShdPnd:\s*([0-9a-f]+)",
                Path(f"/proc/{pid}/status").read_text(),
                flags=re.M,
            ).group(1)
            self.assertEqual(
                sigset_to_hex(task.signal.shared_pending.signal.address_of_()), expected
            )



# Copyright Jason R. Coombs
# SPDX-License-Identifier: MIT
# Based on https://pypi.org/project/jaraco.packaging/.

import os
import subprocess
import sys


def setup(app):
    app.add_config_value("package_url", "", "")
    app.connect("config-inited", load_config_from_setup)
    app.connect("html-page-context", add_package_url)
    return {"parallel_read_safe": "True"}


def load_config_from_setup(app, config):
    """
    Replace values in config from package metadata
    """
    # for now, assume project root is one level up
    root = os.path.join(app.confdir, "..")
    setup_script = os.path.join(root, "setup.py")
    fields = ["--name", "--version", "--url", "--author"]
    dist_info_cmd = [sys.executable, setup_script] + fields
    output = subprocess.check_output(dist_info_cmd, cwd=root, universal_newlines=True)
    outputs = output.strip().split("\n")
    project, version, url, author = outputs
    config.project = project
    config.version = config.release = version
    config.package_url = url
    config.author = config.copyright = author


def add_package_url(app, pagename, templatename, context, doctree):
    context["package_url"] = app.config.package_url



# Copyright (c) 2025, Oracle and/or its affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later


import os
import sys
import tempfile
import traceback
import types

import drgn.cli
from tests import TestCase


class TestCli(TestCase):
    def run_cli(self, args, *, input=None):
        stdout_r, stdout_w = os.pipe()
        stderr_r, stderr_w = os.pipe()
        if input is not None:
            stdin_r, stdin_w = os.pipe()

        pid = os.fork()
        if pid == 0:
            try:
                os.close(stdout_r)
                sys.stdout = open(stdout_w, "w")
                os.close(stderr_r)
                sys.stderr = open(stderr_w, "w")

                if input is not None:
                    os.close(stdin_w)
                    sys.stdin = open(stdin_r, "r")

                sys.argv = ["drgn"] + args

                drgn.cli._main()
            finally:
                exception = sys.exc_info()[1] is not None
                if exception:
                    traceback.print_exc()
                sys.stdout.flush()
                sys.stderr.flush()
                os._exit(1 if exception else 0)

        os.close(stdout_w)
        os.close(stderr_w)

        if input is not None:
            os.close(stdin_r)
            with open(stdin_w, "w") as f:
                f.write(input)

        with open(stdout_r, "r") as f:
            stdout = f.read()
        with open(stderr_r, "r") as f:
            stderr = f.read()

        _, wstatus = os.waitpid(pid, 0)
        if not os.WIFEXITED(wstatus) or os.WEXITSTATUS(wstatus) != 0:
            if os.WIFEXITED(wstatus):
                msg = f"Exited with status {os.WEXITSTATUS(wstatus)}"
            elif os.WIFSIGNALED(wstatus):
                msg = f"Terminated by signal {os.WTERMSIG(wstatus)}"
            else:
                msg = "Exited abnormally"
            self.fail(
                f"""\
{msg}
STDOUT:
{stdout}
STDERR:
{stderr}
"""
            )

        return types.SimpleNamespace(stdout=stdout, stderr=stderr)

    def test_e(self):
        script = r"""
import sys

assert drgn.get_default_prog() is prog
assert __name__ == "__main__"
assert "__file__" not in globals()
assert sys.path[0] == ""
print(sys.argv)
"""
        proc = self.run_cli(
            ["--quiet", "--pid", "0", "--no-default-symbols", "-e", script, "pass"]
        )
        self.assertEqual(proc.stdout, "['-e', 'pass']\n")

    def test_e_empty(self):
        self.run_cli(
            ["--quiet", "--pid", "0", "--no-default-symbols", "-e", ""],
            # This shouldn't be executed.
            input="raise Exception('-e was ignored')",
        )

    def test_script(self):
        with tempfile.NamedTemporaryFile() as f:
            f.write(
                rb"""
assert "drgn" not in globals()

import drgn
import os.path
import sys

assert drgn.get_default_prog() is prog
assert __name__ == "__main__"
assert __file__ == sys.argv[0]
assert sys.path[0] == os.path.dirname(__file__)
print(sys.argv)
"""
            )
            f.flush()
            proc = self.run_cli(
                ["--quiet", "--pid", "0", "--no-default-symbols", f.name, "pass"]
            )
            self.assertEqual(proc.stdout, f"[{f.name!r}, 'pass']\n")

    def test_pipe(self):
        script = r"""
import sys

assert drgn.get_default_prog() is prog
assert __name__ == "__main__"
assert __file__ == "<stdin>"
assert sys.path[0] == ""
# Dummy if statement to test handling of multi-line blocks.
if True:
    print(sys.argv)
"""
        proc = self.run_cli(
            ["--quiet", "--pid", "0", "--no-default-symbols"], input=script
        )
        self.assertEqual(proc.stdout, "['']\n")



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import os
from pathlib import Path
import re
import tempfile

from drgn import Object
from drgn.commands import CommandArgumentError
from drgn.commands._crash.common import CRASH_COMMAND_NAMESPACE
from drgn.helpers.linux.mm import PageUsage
from drgn.helpers.linux.slab import SlabTotalUsage
from tests import with_default_prog
from tests.linux_kernel import (
    HAVE_FULL_MM_SUPPORT,
    possible_cpus,
    skip_unless_have_full_mm_support,
    skip_unless_have_test_kmod,
)
from tests.linux_kernel.crash_commands import CrashCommandTestCase
from tests.linux_kernel.helpers.test_slab import fallback_slab_cache_names
from util import KernelVersion

# Good enough approximation for full kmem -s support.
have_full_kmem_s_support = Path("/proc/slabinfo").exists()


class TestKmem(CrashCommandTestCase):
    def _test_free_common(self, flag):
        cmd = self.check_crash_command(f"kmem -{flag}")
        expected = set(
            re.findall(
                r"^Node\s+([0-9]+)\s*,\s*zone\s+(\w+)",
                Path("/proc/zoneinfo").read_text(),
                flags=re.MULTILINE,
            )
        )
        actual = set()
        header = 0
        for line in cmd.stdout.splitlines():
            if header == 0:
                match = re.match(r"(NODE\s+)?ZONE", line)
                if match:
                    header = 2 if match.group(1) else 1
            else:
                if header == 1:
                    actual.add(("0", re.match(r"\s*[0-9]+\s+(\w+)", line).group(1)))
                else:
                    actual.add(re.match(r"\s*([0-9]+)\s+[0-9]+\s+(\w+)", line).groups())
                header = 0
        # Since Linux kernel commit b2bd8598195f ("mm, vmstat: print
        # non-populated zones in zoneinfo") (in v4.12), these should be equal,
        # but before that, /proc/zoneinfo doesn't include all zones.
        self.assertGreaterEqual(actual, expected)
        self.assertRegex(cmd.stdout, r"nr_free_pages: [0-9]+")

        for variable in (
            "zone",
            "name",
            "size",
            "free",
            "mem_map",
            "start_paddr",
            "start_pfn",
            "order",
            "block_size",
            "migrate_type",
            "num_blocks",
            "num_pages",
        ):
            self.assertIn(variable, cmd.drgn_option.globals)
        for variable in (
            "expected_free_pages",
            "actual_free_pages",
        ):
            self.assertIsInstance(cmd.drgn_option.globals[variable], int)
        if flag == "F":
            self.assertRegex(cmd.stdout, r"(?m)^[0-9a-f]+$")
            self.assertIsInstance(cmd.drgn_option.globals["page"], Object)

    def test_f(self):
        self._test_free_common("f")

    def test_F(self):
        self._test_free_common("F")

    def test_i(self):
        cmd = self.check_crash_command("kmem -i")
        for label in (
            "TOTAL MEM",
            "FREE",
            "USED",
            "BUFFERS",
            "CACHED",
            "SLAB",
            "TOTAL HUGE",
            "HUGE FREE",
            "TOTAL SWAP",
            "SWAP USED",
            "SWAP FREE",
            "COMMIT LIMIT",
            "COMMITTED",
        ):
            self.assertRegex(label, rf"(?m)^{label}\b")
        for variable in (
            "total_mem",
            "free",
            "used",
            "buffers",
            "cached",
            "commit_limit",
            "committed",
        ):
            self.assertIsInstance(cmd.drgn_option.globals[variable], int)
        self.assertIsInstance(cmd.drgn_option.globals["slab_usage"], SlabTotalUsage)
        for variable in (
            "hugetlb_usage",
            "swap_usage",
        ):
            self.assertIsInstance(cmd.drgn_option.globals[variable], PageUsage)

    @skip_unless_have_test_kmod
    def test_v(self):
        cmd = self.check_crash_command("kmem -v")
        self.assertIn(f"{self.prog['drgn_test_vmalloc_va'].value_():x}", cmd.stdout)
        self.assertIn("for_each_vmap_area(", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["va"].type_.type_name(), "struct vmap_area *"
        )
        self.assertEqual(
            cmd.drgn_option.globals["vm"].type_.type_name(), "struct vm_struct *"
        )
        for variable in ("start", "end", "size"):
            self.assertIn(variable, cmd.drgn_option.globals)

    def test_V(self):
        cmd = self.check_crash_command("kmem -V")
        labels = ["VM_ZONE_STAT", "VM_NODE_STAT"]
        vmstat_contents = Path("/proc/vmstat").read_text()
        # Before Linux kernel commit 3a321d2a3dde ("mm: change the call sites
        # of numa statistics items") (in v4.14), NUMA events are zone
        # statistics.
        if KernelVersion(os.uname().release) >= KernelVersion("4.14") and re.search(
            r"^numa_hit\b", vmstat_contents, flags=re.M
        ):
            labels.append(r"VM_NUMA_(EVENT|STAT)")
        if re.search(r"^pgmajfault\b", vmstat_contents, flags=re.M):
            labels.append("VM_EVENT_STATES")
        for label in labels:
            self.assertRegex(cmd.stdout, rf"(?m)^{label}:")

        for helper in (
            "global_zone_page_state",
            "global_node_page_state",
            "global_numa_event_state",
            "global_vm_event_state",
        ):
            self.assertIn(helper, cmd.drgn_option.stdout)
        for variable in ("name", "item", "value"):
            self.assertIn(variable, cmd.drgn_option.globals)

    @skip_unless_have_test_kmod
    def test_n(self):
        cmd = self.check_crash_command("kmem -n")

        expected = set(
            re.findall(
                r"^Node\s+([0-9]+)\s*,\s*zone\s+(\w+)",
                Path("/proc/zoneinfo").read_text(),
                flags=re.MULTILINE,
            )
        )
        actual = set()
        state = None
        for line in cmd.stdout.splitlines():
            if state is None:
                if re.match(r"\s*NODE\b", line):
                    state = "found_node_header"
            elif state == "found_node_header":
                match = re.match(r"\s*([0-9]+)", line)
                if match:
                    node = match.group(1)
                    state = "found_node"
                else:
                    state = None
            elif state == "found_node":
                if re.match(r"\s*ZONE\b", line):
                    state = "found_zone_header"
            elif state == "found_zone_header":
                match = re.match(r"\s*[0-9]+\s*(\w+)", line)
                if match:
                    actual.add((node, match.group(1)))
                else:
                    state = None
        # Since Linux kernel commit b2bd8598195f ("mm, vmstat: print
        # non-populated zones in zoneinfo") (in v4.12), these should be equal,
        # but before that, /proc/zoneinfo doesn't include all zones.
        self.assertGreaterEqual(actual, expected)

        for variable in (
            "node",
            "pgdat",
            "size",
            "start_pfn",
            "mem_map",
            "start_paddr",
            "zone",
            "zone_name",
            "zone_size",
            "zone_start_pfn",
            "zone_mem_map",
            "zone_start_paddr",
        ):
            self.assertIn(variable, cmd.drgn_option.globals)

        # Memory hotplug depends on SPARSEMEM, so this check is close enough
        # for both.
        if Path("/sys/bus/memory").exists():
            expected_section = (
                str(self.prog["drgn_test_section_nr"].value_()),
                f'{self.prog["drgn_test_mem_section"].value_():x}',
            )
            found_section_header = False
            found_section = False
            for line in cmd.stdout.splitlines():
                if found_section_header:
                    match = re.match(r"\s*([0-9]+)\s+([0-9a-f]+)", line)
                    if not match:
                        break
                    if match.groups() == expected_section:
                        found_section = True
                        break
                elif re.match(r"\s*NR\s+SECTION", line):
                    found_section_header = True
            self.assertTrue(found_section)

            expected = os.listdir("/sys/bus/memory/devices")
            actual = []
            found_memory_block_header = False
            for line in cmd.stdout.splitlines():
                if found_memory_block_header:
                    match = re.match(r"\s*[0-9a-f]+\s+(\S+)", line)
                    if not match:
                        break
                    actual.append(match.group(1))
                elif re.match(r"\s*MEMORY_BLOCK\b", line):
                    found_memory_block_header = True
            self.assertCountEqual(actual, expected)

            for variable in (
                "nr",
                "section",
                "coded_mem_map",
                "state",
                "pfn",
                "block_size",
                "mem",
                "name",
                "start_section_no",
                "physical_start",
                "physical_end",
                "node",
            ):
                self.assertIn(variable, cmd.drgn_option.globals)

    def test_z(self):
        cmd = self.check_crash_command("kmem -z")

        expected = set(
            re.findall(
                r"^Node\s+([0-9]+)\s*,\s*zone\s+(\w+)",
                Path("/proc/zoneinfo").read_text(),
                flags=re.MULTILINE,
            )
        )
        actual = set(
            re.findall(
                r'^NODE: ([0-9]+).*NAME: "([^"]+)"',
                cmd.stdout,
                flags=re.MULTILINE,
            )
        )
        # Since Linux kernel commit b2bd8598195f ("mm, vmstat: print
        # non-populated zones in zoneinfo") (in v4.12), these should be equal,
        # but before that, /proc/zoneinfo doesn't include all zones.
        self.assertGreaterEqual(actual, expected)

        for variable in (
            "node",
            "zone",
            "size",
            "present",
            "min_watermark",
            "low_watermark",
            "high_watermark",
            "stat_name",
            "stat_item",
            "stat_value",
        ):
            self.assertIn(variable, cmd.drgn_option.globals)

    def test_o(self):
        cmd = self.check_crash_command("kmem -o")
        for cpu in possible_cpus():
            self.assertRegex(cmd.stdout, rf"CPU {cpu}:\s*[0-9a-f]+")
        self.assertIn("for_each_possible_cpu(", cmd.drgn_option.stdout)
        self.assertIn("per_cpu_ptr(", cmd.drgn_option.stdout)
        self.assertIn("offset", cmd.drgn_option.globals)

    def test_h(self):
        cmd = self.check_crash_command("kmem -h")
        try:
            names = os.listdir("/sys/kernel/mm/hugepages")
        except FileNotFoundError:
            names = []
        for name in names:
            self.assertIn(name, cmd.stdout)
        self.assertIn("for_each_hstate(", cmd.drgn_option.stdout)
        if names:
            self.assertEqual(
                cmd.drgn_option.globals["hstate"].type_.type_name(), "struct hstate *"
            )
            for variable in ("size", "free", "total", "name"):
                self.assertIn(variable, cmd.drgn_option.globals)

    # For kmem -p and kmem -m, printing every page is too slow. Just get the
    # first few.
    @skip_unless_have_full_mm_support
    def test_p(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            path = Path(temp_dir) / "file"
            CRASH_COMMAND_NAMESPACE.run(self.prog, f"kmem -p | head > {path}")
            for line in path.read_text().splitlines()[1:]:
                self.assertRegex(
                    line,
                    # PAGE
                    r"^[0-9a-f]+\s+"
                    # PHYSICAL
                    r"[0-9a-f]+\s+"
                    # MAPPING
                    r"[0-9a-f]+\s+"
                    # INDEX
                    r"[0-9a-f]+\s+"
                    # CNT
                    r"-?[0-9]+\s+"
                    # FLAGS
                    r"[0-9a-f]+( [\w,]+)?$",
                )

        drgn_option = self.run_crash_command_drgn_option("kmem -p", mode="capture")
        drgn_option_globals = {"prog": self.prog}
        with with_default_prog(self.prog):
            exec(drgn_option.stdout + "\n    break", drgn_option_globals)
        for variable in ("physical", "mapping", "index", "cnt", "flags"):
            self.assertIsInstance(drgn_option_globals[variable], Object)
        self.assertIsInstance(drgn_option_globals["decoded_flags"], str)

    @skip_unless_have_full_mm_support
    def test_m(self):
        members = "mapping,private,_refcount,lru,flags"
        with tempfile.TemporaryDirectory() as temp_dir:
            path = Path(temp_dir) / "file"
            CRASH_COMMAND_NAMESPACE.run(self.prog, f"kmem -m {members} | head > {path}")
            for line in path.read_text().splitlines()[1:]:
                self.assertRegex(
                    line,
                    # PAGE
                    r"^[0-9a-f]+\s+"
                    # mapping
                    r"[0-9a-f]+\s+"
                    # private
                    r"-?[0-9]+\s+"
                    # _refcount
                    r"-?[0-9]+\s+"
                    # lru
                    r"[0-9a-f]+,[0-9a-f]+\s+"
                    # flags
                    r"[0-9a-f]+$",
                )

        drgn_option = self.run_crash_command_drgn_option(
            f"kmem -m {members}", mode="capture"
        )
        drgn_option_globals = {"prog": self.prog}
        with with_default_prog(self.prog):
            exec(drgn_option.stdout + "\n    break", drgn_option_globals)
        for variable in ("mapping", "private", "_refcount", "lru", "flags"):
            self.assertIsInstance(drgn_option_globals[variable], Object)

    def check_kmem_s(self, options, check_common=True):
        cmd = self.check_crash_command("kmem -s " + options)
        self.assertEqual(
            cmd.drgn_option.globals["cache"].type_.type_name(), "struct kmem_cache *"
        )
        if check_common:
            for variable in (
                "objsize",
                "name",
            ):
                self.assertIn(variable, cmd.drgn_option.globals)
            if have_full_kmem_s_support:
                for variable in (
                    "usage",
                    "allocated",
                    "total",
                    "slabs",
                    "ssize",
                    "name",
                ):
                    self.assertIn(variable, cmd.drgn_option.globals)
        return cmd

    def test_s(self):
        cmd = self.check_kmem_s("")

        for name in fallback_slab_cache_names(self.prog):
            if have_full_kmem_s_support:
                self.assertRegex(cmd.stdout, rf"[0-9]+k\s+{re.escape(name.decode())}\b")
            else:
                self.assertRegex(cmd.stdout, rf"\b{re.escape(name.decode())}\b")

        self.assertIn("for_each_slab_cache(", cmd.drgn_option.stdout)

    def test_s_match_one(self):
        names = sorted(name.decode() for name in fallback_slab_cache_names(self.prog))
        cmd = self.check_kmem_s(names[0])

        if have_full_kmem_s_support:
            self.assertRegex(cmd.stdout, rf"[0-9]+k\s+{re.escape(names[0])}\b")
        else:
            self.assertRegex(cmd.stdout, rf"\b{re.escape(names[0])}\b")
        self.assertNotRegex(cmd.stdout, rf"\b{re.escape(names[1])}\b")

        self.assertIn('find_slab_cache("', cmd.drgn_option.stdout)

    def test_s_match_multiple(self):
        names = sorted(name.decode() for name in fallback_slab_cache_names(self.prog))
        cmd = self.check_kmem_s(" ".join(names))

        for name in names:
            if have_full_kmem_s_support:
                self.assertRegex(cmd.stdout, rf"[0-9]+k\s+{re.escape(name)}\b")
            else:
                self.assertRegex(cmd.stdout, rf"\b{re.escape(name)}\b")

        self.assertIn("find_slab_cache(search_name)", cmd.drgn_option.stdout)

    def test_s_ignore_one(self):
        ignore = min(fallback_slab_cache_names(self.prog)).decode()
        cmd = self.check_kmem_s(f"-I {ignore}")

        self.assertRegex(cmd.stdout, rf"\[IGNORED\].*\b{re.escape(ignore)}\b")

        self.assertIn("for_each_slab_cache(", cmd.drgn_option.stdout)
        self.assertRegex(cmd.drgn_option.stdout, rf"== .*\b{re.escape(ignore)}\b")

    def test_s_ignore_multiple(self):
        names = sorted(name.decode() for name in fallback_slab_cache_names(self.prog))
        cmd = self.check_kmem_s(f"-I {','.join(names)}", check_common=False)

        for name in names:
            self.assertRegex(cmd.stdout, rf"\[IGNORED\].*\b{re.escape(name)}\b")

        self.assertIn("for_each_slab_cache(", cmd.drgn_option.stdout)
        self.assertIn(" in ignore:", cmd.drgn_option.stdout)

    def test_s_match_and_ignore(self):
        names = sorted(name.decode() for name in fallback_slab_cache_names(self.prog))
        cmd = self.check_kmem_s(f"{names[0]} -I {names[0]}", check_common=False)

        self.assertRegex(cmd.stdout, rf"\[IGNORED\].*\b{re.escape(names[0])}\b")
        self.assertNotRegex(cmd.stdout, rf"\b{re.escape(names[1])}\b")

        self.assertIn("find_slab_cache(", cmd.drgn_option.stdout)
        self.assertRegex(cmd.drgn_option.stdout, rf"== .*\b{re.escape(names[0])}\b")

    def test_g_value(self):
        value = (1 << self.prog["PG_locked"].value_()) | (
            1 << self.prog["PG_uptodate"].value_()
        )
        cmd = self.check_crash_command(f"kmem -g {value:x}")
        self.assertIn("FLAGS:", cmd.stdout)
        self.assertIn("PG_locked", cmd.stdout)
        self.assertIn("PG_uptodate", cmd.stdout)
        self.assertNotIn("PG_dirty", cmd.stdout)
        self.assertIn("decode_page_flags_value", cmd.drgn_option.stdout)
        self.assertIn("PG_locked", cmd.drgn_option.globals["flags"])

    def test_g_no_value(self):
        cmd = self.check_crash_command("kmem -g")
        self.assertIn("PG_uptodate", cmd.stdout)
        self.assertIn(".enumerators", cmd.drgn_option.stdout)
        for variable in ("name", "bit", "value"):
            self.assertIn(variable, cmd.drgn_option.globals)

    @skip_unless_have_test_kmod
    def test_identify_symbol(self):
        address = self.prog.symbol("drgn_test_function").address
        cmd = self.run_crash_command(f"kmem {address:x}")
        self.assertRegex(
            cmd.stdout, rf"(?m)^{address:x} \(.\) drgn_test_function \[drgn_test\]$"
        )

    @skip_unless_have_test_kmod
    def test_identify_symbol_offset(self):
        symbol = self.prog.symbol("drgn_test_function")
        address = symbol.address + symbol.size - 1
        cmd = self.run_crash_command(f"kmem {address:x}")
        self.assertRegex(
            cmd.stdout,
            rf"(?m)^{address:x} \(.\) drgn_test_function\+{symbol.size - 1} \[drgn_test\]$",
        )

    def test_identify_task(self):
        cmd = self.check_crash_command(f"kmem {self.prog['init_task'].address_:x}")
        self.assertRegex(cmd.stdout, r'(?m)^COMMAND: "swapper')
        self.assertIn("identify_address", cmd.drgn_option.globals)
        self.assertTrue(cmd.drgn_option.globals["identified"])

    def test_identify_task_stack(self):
        cmd = self.run_crash_command(f"kmem {self.prog['init_task'].stack.value_():x}")
        self.assertRegex(cmd.stdout, r'(?m)^COMMAND: "swapper')

    @skip_unless_have_test_kmod
    def test_identify_vmalloc(self):
        address = self.prog["drgn_test_vmalloc_va"].value_()
        for option in ("", " -v"):
            with self.subTest(option=option):
                cmd = self.run_crash_command(f"kmem{option} {address:x}")
                self.assertIn("VMAP_AREA", cmd.stdout)
                self.assertIn(f"{address:x}", cmd.stdout)
                page_address = f"{self.prog['drgn_test_vmalloc_page'].value_():x}"
                if HAVE_FULL_MM_SUPPORT:
                    if option:
                        self.assertNotIn(page_address, cmd.stdout)
                    else:
                        self.assertIn(page_address, cmd.stdout)

    @skip_unless_have_test_kmod
    def test_identify_not_vmalloc(self):
        address = self.prog["drgn_test_small_slab_objects"][0].value_()
        cmd = self.run_crash_command(f"kmem -v {address:x}")
        self.assertIn("address is not allocated in vmalloc subsystem", cmd.stdout)
        self.assertNotIn("SLAB", cmd.stdout)

    @skip_unless_have_test_kmod
    def test_identify_page(self):
        address = self.prog["drgn_test_page"].value_()
        for option in ("", " -p"):
            with self.subTest(option=option):
                cmd = self.run_crash_command(f"kmem{option} {address + 1:x}")
                self.assertIn("PAGE", cmd.stdout)
                self.assertIn(f"{address:x}", cmd.stdout)

    @skip_unless_have_full_mm_support
    @skip_unless_have_test_kmod
    def test_identify_slab(self):
        address = self.prog["drgn_test_small_slab_objects"][0].value_()
        for option in ("", " -s"):
            with self.subTest(option=option):
                cmd = self.run_crash_command(f"kmem{option} {address + 1:x}")
                if self.prog["drgn_test_slob"]:
                    self.assertIn(
                        f"kmem: address is from SLOB: {address + 1:x}", cmd.stdout
                    )
                else:
                    self.assertIn("drgn_test_small", cmd.stdout)
                    self.assertIn(f"[{address:x}]", cmd.stdout)

    @skip_unless_have_test_kmod
    def test_identify_not_slab(self):
        address = self.prog["drgn_test_vmalloc_va"].value_()
        cmd = self.run_crash_command(f"kmem -s {address:x}")
        self.assertIn("address is not allocated in slab subsystem", cmd.stdout)
        self.assertNotIn("VMAP_AREA", cmd.stdout)

    @skip_unless_have_full_mm_support
    @skip_unless_have_test_kmod
    def test_identify_slab_with_names(self):
        address = self.prog["drgn_test_small_slab_objects"][0].value_()
        cmd = self.run_crash_command(f"kmem -s drgn_test_big {address + 1:x}")
        self.assertIn("ignoring pre-selected slab caches for address", cmd.stdout)
        if self.prog["drgn_test_slob"]:
            self.assertIn(f"kmem: address is from SLOB: {address + 1:x}", cmd.stdout)
        else:
            self.assertIn("drgn_test_small", cmd.stdout)
            self.assertIn(f"[{address:x}]", cmd.stdout)

    @skip_unless_have_full_mm_support
    @skip_unless_have_test_kmod
    def test_identify_slab_ignored(self):
        address = self.prog["drgn_test_small_slab_objects"][0].value_()
        cmd = self.run_crash_command(f"kmem -s -I drgn_test_small {address + 1:x}")
        if self.prog["drgn_test_slob"]:
            self.assertIn(f"kmem: address is from SLOB: {address + 1:x}", cmd.stdout)
        else:
            self.assertIn("drgn_test_small", cmd.stdout)
            self.assertIn("[IGNORED]", cmd.stdout)
            self.assertNotIn(f"[{address:x}]", cmd.stdout)

    @skip_unless_have_full_mm_support
    @skip_unless_have_test_kmod
    def test_identify_multiple(self):
        slab_address = self.prog["drgn_test_small_slab_objects"][0].value_()
        vmalloc_address = self.prog["drgn_test_vmalloc_va"].value_()
        cmd = self.check_crash_command(f"kmem {vmalloc_address:x} {slab_address:x}")

        self.assertIn("VMAP_AREA", cmd.stdout)
        self.assertIn(f"{vmalloc_address:x}", cmd.stdout)

        if not self.prog["drgn_test_slob"]:
            self.assertIn("drgn_test_small", cmd.stdout)
            self.assertIn(f"[{slab_address:x}]", cmd.stdout)

        self.assertIn("identify_address", cmd.drgn_option.globals)
        self.assertTrue(cmd.drgn_option.globals["identified"])

    def test_no_arguments(self):
        self.assertRaises(CommandArgumentError, self.run_crash_command, "kmem")



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later


from drgn import NoDefaultProgramError, TypeMember
from drgn.helpers.common.type import member_at_offset, typeof_member
from tests import TestCase, mock_program, with_default_prog


class TestMemberAtOffset(TestCase):
    def test_simple(self):
        prog = mock_program()
        point_type = prog.struct_type(
            "point",
            8,
            (
                TypeMember(prog.int_type("int", 4, True), "x", 0),
                TypeMember(prog.int_type("int", 4, True), "y", 32),
            ),
        )
        self.assertEqual(member_at_offset(point_type, 0), "x")
        self.assertEqual(member_at_offset(point_type, 4), "y")

    def test_unnamed_member(self):
        prog = mock_program()
        type = prog.struct_type(
            None,
            8,
            (
                TypeMember(
                    prog.struct_type(
                        None,
                        8,
                        (
                            TypeMember(prog.int_type("int", 4, True), "x", 0),
                            TypeMember(prog.int_type("int", 4, True), "y", 32),
                        ),
                    ),
                    None,
                ),
            ),
        )
        self.assertEqual(member_at_offset(type, 0), "x")
        self.assertEqual(member_at_offset(type, 4), "y")

    def test_end(self):
        prog = mock_program()
        point_type = prog.struct_type(
            "point",
            8,
            (
                TypeMember(prog.int_type("int", 4, True), "x", 0),
                TypeMember(prog.int_type("int", 4, True), "y", 32),
            ),
        )
        self.assertEqual(member_at_offset(point_type, 8), "<end>")
        self.assertEqual(member_at_offset(point_type, 9), "<past end>")

    def test_padding_between_members(self):
        prog = mock_program()
        type = prog.struct_type(
            None,
            16,
            (
                TypeMember(prog.int_type("int", 4, True), "i", 0),
                TypeMember(prog.int_type("long", 8, True), "l", 64),
            ),
        )
        self.assertEqual(member_at_offset(type, 4), "<padding between i and l>")

    def test_padding_between_unnamed_members(self):
        prog = mock_program()

        type = prog.struct_type(
            None,
            16,
            (
                TypeMember(prog.int_type("int", 4, True), "i"),
                TypeMember(
                    prog.struct_type(
                        None, 8, (TypeMember(prog.int_type("long", 8, True), "l"),)
                    ),
                    None,
                    64,
                ),
            ),
        )
        self.assertEqual(member_at_offset(type, 4), "<padding between i and <unnamed>>")

        type = prog.struct_type(
            None,
            16,
            (
                TypeMember(
                    prog.struct_type(
                        None, 4, (TypeMember(prog.int_type("int", 4, True), "i"),)
                    ),
                    None,
                ),
                TypeMember(prog.int_type("long", 8, True), "l", 64),
            ),
        )
        self.assertEqual(member_at_offset(type, 4), "<padding between <unnamed> and l>")

        type = prog.struct_type(
            None,
            16,
            (
                TypeMember(
                    prog.struct_type(
                        None, 4, (TypeMember(prog.int_type("int", 4, True), "i"),)
                    ),
                    None,
                ),
                TypeMember(
                    prog.struct_type(
                        None, 8, (TypeMember(prog.int_type("long", 8, True), "l"),)
                    ),
                    None,
                    64,
                ),
            ),
        )
        self.assertEqual(
            member_at_offset(type, 4), "<padding between <unnamed> and <unnamed>>"
        )

    def test_padding_at_end(self):
        prog = mock_program()
        type = prog.struct_type(
            None,
            16,
            (
                TypeMember(prog.int_type("long", 8, True), "l", 0),
                TypeMember(prog.int_type("int", 4, True), "i", 64),
            ),
        )
        self.assertEqual(member_at_offset(type, 12), "<padding at end>")

    def test_padding_at_beginning(self):
        prog = mock_program()
        type = prog.struct_type(
            None,
            16,
            (TypeMember(prog.int_type("int", 4, True), "i", 32),),
        )
        self.assertEqual(member_at_offset(type, 0), "<padding at beginning>")

    def test_offset(self):
        prog = mock_program()
        point_type = prog.struct_type(
            "point",
            8,
            (
                TypeMember(prog.int_type("int", 4, True), "x", 0),
                TypeMember(prog.int_type("int", 4, True), "y", 32),
            ),
        )
        self.assertEqual(member_at_offset(point_type, 2), "x+0x2")
        self.assertEqual(member_at_offset(point_type, 6), "y+0x2")

    def test_nested(self):
        prog = mock_program()
        point_type = prog.struct_type(
            "point",
            8,
            (
                TypeMember(prog.int_type("int", 4, True), "x", 0),
                TypeMember(prog.int_type("int", 4, True), "y", 32),
            ),
        )
        line_segment_type = prog.struct_type(
            "line_segment",
            16,
            (TypeMember(point_type, "a"), TypeMember(point_type, "b", 64)),
        )
        self.assertEqual(member_at_offset(line_segment_type, 0), "a.x")
        self.assertEqual(member_at_offset(line_segment_type, 4), "a.y")
        self.assertEqual(member_at_offset(line_segment_type, 8), "b.x")
        self.assertEqual(member_at_offset(line_segment_type, 12), "b.y")

    def test_padding_in_nested(self):
        prog = mock_program()
        type = prog.struct_type(
            None,
            16,
            (
                TypeMember(
                    prog.struct_type(
                        None,
                        16,
                        (
                            TypeMember(prog.int_type("int", 4, True), "i", 0),
                            TypeMember(prog.int_type("long", 8, True), "l", 64),
                        ),
                    ),
                    "x",
                ),
            ),
        )
        self.assertEqual(member_at_offset(type, 4), "x.<padding between i and l>")

    def test_array(self):
        prog = mock_program()
        type = prog.array_type(prog.int_type("int", 4, True), 2)
        self.assertEqual(member_at_offset(type, 0), "[0]")
        self.assertEqual(member_at_offset(type, 4), "[1]")

    def test_array_end(self):
        prog = mock_program()
        type = prog.array_type(prog.int_type("int", 4, True), 2)
        self.assertEqual(member_at_offset(type, 8), "<end>")
        self.assertEqual(member_at_offset(type, 9), "<past end>")

    def test_array_padding_at_end_of_element(self):
        prog = mock_program()
        type = prog.array_type(
            prog.struct_type(
                None,
                16,
                (
                    TypeMember(prog.int_type("long", 8, True), "l", 0),
                    TypeMember(prog.int_type("int", 4, True), "i", 64),
                ),
            ),
            2,
        )
        self.assertEqual(member_at_offset(type, 12), "[0].<padding at end>")

    def test_array_offset(self):
        prog = mock_program()
        type = prog.array_type(prog.int_type("int", 4, True), 2)
        self.assertEqual(member_at_offset(type, 1), "[0]+0x1")
        self.assertEqual(member_at_offset(type, 7), "[1]+0x3")

    def test_array_of_structs(self):
        prog = mock_program()
        type = prog.array_type(
            prog.struct_type(
                "point",
                8,
                (
                    TypeMember(prog.int_type("int", 4, True), "x", 0),
                    TypeMember(prog.int_type("int", 4, True), "y", 32),
                ),
            ),
            2,
        )
        self.assertEqual(member_at_offset(type, 0), "[0].x")
        self.assertEqual(member_at_offset(type, 4), "[0].y")
        self.assertEqual(member_at_offset(type, 8), "[1].x")
        self.assertEqual(member_at_offset(type, 12), "[1].y")

    def test_array_of_structs_end(self):
        prog = mock_program()
        type = prog.array_type(
            prog.struct_type(
                "point",
                8,
                (
                    TypeMember(prog.int_type("int", 4, True), "x", 0),
                    TypeMember(prog.int_type("int", 4, True), "y", 32),
                ),
            ),
            2,
        )
        self.assertEqual(member_at_offset(type, 16), "<end>")
        self.assertEqual(member_at_offset(type, 17), "<past end>")

    def test_array_of_structs_padding(self):
        prog = mock_program()
        type = prog.array_type(
            prog.struct_type(
                None,
                16,
                (
                    TypeMember(prog.int_type("int", 4, True), "i", 0),
                    TypeMember(prog.int_type("long", 8, True), "l", 64),
                ),
            ),
            2,
        )
        self.assertEqual(member_at_offset(type, 4), "[0].<padding between i and l>")
        self.assertEqual(member_at_offset(type, 20), "[1].<padding between i and l>")

    def test_array_member(self):
        prog = mock_program()
        type = prog.struct_type(
            None,
            16,
            (
                TypeMember(prog.array_type(prog.int_type("int", 4, True), 3), "a", 0),
                TypeMember(prog.array_type(prog.int_type("int", 4, True), 1), "b", 96),
            ),
        )
        self.assertEqual(member_at_offset(type, 0), "a[0]")
        self.assertEqual(member_at_offset(type, 4), "a[1]")
        self.assertEqual(member_at_offset(type, 8), "a[2]")
        self.assertEqual(member_at_offset(type, 12), "b[0]")

    def test_array_member_end(self):
        prog = mock_program()
        type = prog.struct_type(
            None,
            16,
            (
                TypeMember(prog.array_type(prog.int_type("int", 4, True), 3), "a", 0),
                TypeMember(prog.array_type(prog.int_type("int", 4, True), 1), "b", 96),
            ),
        )
        self.assertEqual(member_at_offset(type, 16), "<end>")
        self.assertEqual(member_at_offset(type, 17), "<past end>")

    def test_incomplete_array(self):
        prog = mock_program()
        type = prog.array_type(prog.int_type("int", 4, True), None)
        self.assertEqual(member_at_offset(type, 0), "[0]")
        self.assertEqual(member_at_offset(type, 8), "[2]")

    def test_incomplete_array_offset(self):
        prog = mock_program()
        type = prog.array_type(prog.int_type("int", 4, True), None)
        self.assertEqual(member_at_offset(type, 3), "[0]+0x3")
        self.assertEqual(member_at_offset(type, 10), "[2]+0x2")

    def test_incomplete_array_member(self):
        prog = mock_program()
        type = prog.struct_type(
            None,
            16,
            (
                TypeMember(prog.int_type("int", 4, True), "n", 0),
                TypeMember(
                    prog.array_type(prog.int_type("int", 4, True), None), "a", 32
                ),
            ),
        )
        self.assertEqual(member_at_offset(type, 0), "n")
        self.assertEqual(member_at_offset(type, 4), "a[0]")
        self.assertEqual(member_at_offset(type, 8), "a[1]")

    def test_union(self):
        prog = mock_program()
        option_type = prog.union_type(
            "option",
            4,
            (
                TypeMember(prog.int_type("int", 4, True), "i"),
                TypeMember(prog.float_type("float", 4), "f"),
            ),
        )
        self.assertEqual(member_at_offset(option_type, 0), "i or f")

    def test_union_end(self):
        prog = mock_program()
        option_type = prog.union_type(
            "option",
            4,
            (
                TypeMember(prog.int_type("int", 4, True), "i"),
                TypeMember(prog.float_type("float", 4), "f"),
            ),
        )
        self.assertEqual(member_at_offset(option_type, 4), "<end>")

    def test_union_of_structs(self):
        prog = mock_program()
        type = prog.union_type(
            None,
            8,
            (
                TypeMember(
                    prog.struct_type(
                        None,
                        8,
                        (
                            TypeMember(prog.int_type("int", 4, True), "x", 0),
                            TypeMember(prog.int_type("int", 4, True), "y", 32),
                        ),
                    ),
                    "point",
                ),
                TypeMember(prog.float_type("int", 4), "i"),
            ),
        )
        self.assertEqual(member_at_offset(type, 0), "point.x or i")
        self.assertEqual(member_at_offset(type, 4), "point.y")

    def test_unions_nested(self):
        prog = mock_program()
        type = prog.union_type(
            None,
            8,
            (
                TypeMember(
                    prog.union_type(
                        None,
                        8,
                        (
                            TypeMember(prog.int_type("long", 8, True), "l"),
                            TypeMember(prog.float_type("double", 8), "d"),
                        ),
                    ),
                    "big",
                ),
                TypeMember(
                    prog.union_type(
                        None,
                        4,
                        (
                            TypeMember(prog.int_type("int", 4, True), "i"),
                            TypeMember(prog.float_type("float", 4), "f"),
                        ),
                    ),
                    "small",
                ),
            ),
        )
        self.assertEqual(
            member_at_offset(type, 0), "big.l or big.d or small.i or small.f"
        )
        self.assertEqual(member_at_offset(type, 4), "big.l+0x4 or big.d+0x4")

    def test_type_error(self):
        prog = mock_program()
        self.assertRaises(TypeError, member_at_offset, prog.int_type("int", 4, True), 0)

    def test_typedefs(self):
        prog = mock_program()
        point_type = prog.typedef_type(
            "Point",
            prog.struct_type(
                None,
                8,
                (
                    TypeMember(prog.int_type("int", 4, True), "x", 0),
                    TypeMember(prog.int_type("int", 4, True), "y", 32),
                ),
            ),
        )
        line_segment_type = prog.typedef_type(
            "LineSegment",
            prog.struct_type(
                None,
                16,
                (TypeMember(point_type, "a"), TypeMember(point_type, "b", 64)),
            ),
        )
        self.assertEqual(member_at_offset(line_segment_type, 0), "a.x")
        self.assertEqual(member_at_offset(line_segment_type, 4), "a.y")
        self.assertEqual(member_at_offset(line_segment_type, 8), "b.x")
        self.assertEqual(member_at_offset(line_segment_type, 12), "b.y")

    def test_default_prog(self):
        self.assertRaises(NoDefaultProgramError, member_at_offset, "struct point", 0)

        types = []
        prog = mock_program(types=types)
        types.append(
            prog.struct_type(
                "point",
                8,
                (
                    TypeMember(prog.int_type("int", 4, True), "x", 0),
                    TypeMember(prog.int_type("int", 4, True), "y", 32),
                ),
            )
        )
        with with_default_prog(prog):
            self.assertIdentical(member_at_offset("struct point", 0), "x")
            self.assertRaises(LookupError, member_at_offset, "struct foo", 0)


class TestTypeofMember(TestCase):
    def test_simple(self):
        prog = mock_program()
        point_type = prog.struct_type(
            "point",
            8,
            (
                TypeMember(prog.int_type("int", 4, True), "x", 0),
                TypeMember(prog.int_type("int", 4, True), "y", 32),
            ),
        )
        self.assertIdentical(
            typeof_member(point_type, "x"), prog.int_type("int", 4, True)
        )

    def test_default_prog(self):
        self.assertRaises(NoDefaultProgramError, typeof_member, "struct point", "x")

        types = []
        prog = mock_program(types=types)
        types.append(
            prog.struct_type(
                "point",
                8,
                (
                    TypeMember(prog.int_type("int", 4, True), "x", 0),
                    TypeMember(prog.int_type("int", 4, True), "y", 32),
                ),
            )
        )
        with with_default_prog(prog):
            self.assertIdentical(
                typeof_member("struct point", "x"), prog.int_type("int", 4, True)
            )
            self.assertRaises(LookupError, typeof_member, "struct foo", "x")



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import os

from drgn import Object, ObjectNotFoundError, Program
from tests.linux_kernel import LinuxKernelTestCase, skip_unless_have_test_kmod


class TestJiffies(LinuxKernelTestCase):
    def test_jiffies(self):
        self.assertIdentical(
            self.prog["jiffies"],
            Object(
                self.prog,
                "volatile unsigned long",
                address=self.prog.symbol("jiffies").address,
            ),
        )


class TestUts(LinuxKernelTestCase):
    def test_uts_release(self):
        self.assertEqual(
            self.prog["UTS_RELEASE"].string_().decode(), os.uname().release
        )

    def test_uts_release_no_debug_info(self):
        prog = Program()
        prog.set_kernel()
        self.assertEqual(prog["UTS_RELEASE"].string_().decode(), os.uname().release)


class TestVmcoreinfo(LinuxKernelTestCase):
    def test_vmcoreinfo(self):
        vmcoreinfo_data = dict(
            line.split("=", 1)
            for line in self.prog["VMCOREINFO"].string_().decode().strip().split("\n")
        )
        self.assertEqual(
            int(vmcoreinfo_data["SYMBOL(init_uts_ns)"], 16),
            self.prog.symbol("init_uts_ns").address,
        )

    def test_vmcoreinfo_no_debug_info(self):
        prog = Program()
        prog.set_kernel()
        vmcoreinfo_data = dict(
            line.split("=", 1)
            for line in prog["VMCOREINFO"].string_().decode().strip().split("\n")
        )
        self.assertEqual(
            vmcoreinfo_data["OSRELEASE"],
            os.uname().release,
        )

    @skip_unless_have_test_kmod
    def test_constants(self):
        for constant in (
            "THREAD_SIZE",
            "NR_SECTION_ROOTS",
            "SECTIONS_PER_ROOT",
            "SECTION_SIZE_BITS",
            "MAX_PHYSMEM_BITS",
            "ARCH_PFN_OFFSET",
        ):
            with self.subTest(constant=constant):
                try:
                    expected = self.prog["drgn_test_" + constant].read_()
                except ObjectNotFoundError:
                    with self.assertRaises(ObjectNotFoundError):
                        self.prog[constant]
                else:
                    self.assertEqual(self.prog[constant], expected)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import os

from drgn import TypeKind
from drgn.helpers.linux.device import (
    MAJOR,
    bus_for_each_dev,
    class_for_each_device,
    dev_name,
    for_each_registered_blkdev,
    for_each_registered_chrdev,
)
from tests.linux_kernel import LinuxKernelTestCase


class TestDevice(LinuxKernelTestCase):
    def test_bus_for_each_dev(self):
        # This also tests dev_name() and bus_to_subsys().
        self.assertCountEqual(
            [
                dev_name(dev)
                for dev in bus_for_each_dev(self.prog["cpu_subsys"].address_of_())
            ],
            os.listdir(b"/sys/bus/cpu/devices"),
        )

    def test_class_for_each_device(self):
        # Before Linux kernel commit 7671284b6c77 ("/dev/mem: make mem_class a
        # static const structure") (in v6.5), mem_class is a pointer instead of
        # a struct.
        mem_class = self.prog["mem_class"]
        if mem_class.type_.unaliased_kind() != TypeKind.POINTER:
            mem_class = mem_class.address_of_()
        # This also tests dev_name() and class_to_subsys().
        self.assertCountEqual(
            [dev_name(dev) for dev in class_for_each_device(mem_class)],
            os.listdir(b"/sys/class/mem"),
        )

    def test_for_each_registered_chrdev(self):
        expected = []
        with open("/proc/devices", "rb") as f:
            ignore = True
            for line in f:
                line = line.rstrip(b"\n")
                if line == b"Character devices:":
                    ignore = False
                elif ignore:
                    pass
                elif not line:
                    break
                else:
                    tokens = line.split(maxsplit=1)
                    expected.append((int(tokens[0]), tokens[1]))

        actual = []
        for dev, _, name, cdev in for_each_registered_chrdev(self.prog):
            actual.append((MAJOR(dev), name))
            self.assertEqual(cdev.type_.type_name(), "struct cdev *")

        self.assertCountEqual(actual, expected)

    def test_for_each_registered_blkdev(self):
        expected = []
        with open("/proc/devices", "rb") as f:
            ignore = True
            for line in f:
                line = line.rstrip(b"\n")
                if line == b"Block devices:":
                    ignore = False
                elif ignore:
                    pass
                elif not line:
                    break
                else:
                    tokens = line.split(maxsplit=1)
                    expected.append((int(tokens[0]), tokens[1]))

        actual = []
        for major, name, obj in for_each_registered_blkdev(self.prog):
            actual.append((major, name))
            self.assertEqual(obj.type_.type_name(), "struct blk_major_name *")

        self.assertCountEqual(actual, expected)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import logging
import os
import re
import unittest

from _drgn_util.platform import NORMALIZED_MACHINE_NAME
from drgn import Object, Program, TypeMember, reinterpret
from drgn.helpers.linux import load_module_kallsyms, load_vmlinux_kallsyms
from tests import assertReprPrettyEqualsStr, drgn_log_level, modifyenv
from tests.linux_kernel import (
    LinuxKernelTestCase,
    fork_and_stop,
    skip_unless_have_stack_tracing,
    skip_unless_have_test_kmod,
)


@skip_unless_have_stack_tracing
class LinuxKernelStackTraceTestCase(LinuxKernelTestCase):
    def _test_drgn_test_kthread_trace(self, trace, test_source=True):
        for i, frame in enumerate(trace):
            if frame.name == "drgn_test_kthread_fn3":
                break
        else:
            self.fail("Couldn't find drgn_test_kthread_fn3 frame")
        self.assertEqual(trace[i + 1].name, "drgn_test_kthread_fn2")
        self.assertEqual(trace[i + 2].name, "drgn_test_kthread_fn")

        if test_source:
            source = frame.source()
            self.assertRegex(source.filename, r"\bdrgn_test\.c$")
            self.assertIsInstance(source.line, int)
            self.assertIsInstance(source.column, int)
            self.assertEqual(
                tuple(source), (source.filename, source.line, source.column)
            )
            self.assertEqual(source.name(), "drgn_test_kthread_fn3")


class TestStackTrace(LinuxKernelStackTraceTestCase):
    @skip_unless_have_test_kmod
    def test_by_task_struct(self):
        self._test_drgn_test_kthread_trace(
            self.prog.stack_trace(self.prog["drgn_test_kthread"])
        )

    def _test_by_pid(self, orc):
        old_orc = int(os.environ.get("DRGN_PREFER_ORC_UNWINDER", "0")) != 0
        with modifyenv({"DRGN_PREFER_ORC_UNWINDER": "1" if orc else "0"}):
            if orc == old_orc:
                prog = self.prog
            else:
                prog = Program()
                prog.set_kernel()
                self._load_debug_info(prog)
            self._test_drgn_test_kthread_trace(
                prog.stack_trace(prog["drgn_test_kthread"].pid)
            )

    @skip_unless_have_test_kmod
    def test_by_pid_dwarf(self):
        self._test_by_pid(False)

    @unittest.skipUnless(
        NORMALIZED_MACHINE_NAME == "x86_64",
        f"{NORMALIZED_MACHINE_NAME} does not use ORC",
    )
    @skip_unless_have_test_kmod
    def test_by_pid_orc(self):
        self._test_by_pid(True)

    def _check_logged_orc_message(self, captured_logs, module):
        # To be sure that we actually used ORC to unwind through the drgn_test
        # stack frames, search for the log output. We don't know which ORC
        # version is used, so just ensure that we have a log line that mentions
        # loading ORC.
        expr = re.compile(
            r"DEBUG:drgn:Loaded built-in ORC \(v\d+\) for module " + module
        )
        for line in captured_logs.output:
            if expr.fullmatch(line):
                break
        else:
            self.fail(f"Did not load built-in ORC for {module}")

    @unittest.skipUnless(
        NORMALIZED_MACHINE_NAME == "x86_64",
        f"{NORMALIZED_MACHINE_NAME} does not use ORC",
    )
    @skip_unless_have_test_kmod
    def test_by_pid_builtin_orc(self):
        # ORC was introduced in kernel 4.14. Detect the presence of ORC or skip
        # the test.
        try:
            self.prog.symbol("__start_orc_unwind")
        except LookupError:
            ver = self.prog["UTS_RELEASE"].string_().decode()
            self.skipTest(f"ORC is not available for {ver}")

        with drgn_log_level(logging.DEBUG):
            # Create a program with the core kernel debuginfo loaded,
            # but without module debuginfo. Load a symbol finder using
            # kallsyms so that the module's stack traces can still have
            # usable frame names.
            prog = Program()
            prog.set_kernel()
            prog.load_debug_info(main=True)
            # Now that vmlinux is loaded, enumerate all the kernel modules so
            # that a drgn_module is created to hold the ORC data
            prog.create_loaded_modules()
            kallsyms = load_module_kallsyms(prog)
            prog.register_symbol_finder("module_kallsyms", kallsyms, enable_index=1)
            for thread in prog.threads():
                if b"drgn_test_kthread".startswith(thread.object.comm.string_()):
                    pid = thread.tid
                    break
            else:
                self.fail("couldn't find drgn_test_kthread")
            # We must set drgn's log level manually, beacuse it won't log messages
            # to the logger if it isn't enabled for them.
            with self.assertLogs("drgn", logging.DEBUG) as log:
                self._test_drgn_test_kthread_trace(
                    prog.stack_trace(pid), test_source=False
                )

            self._check_logged_orc_message(log, "drgn_test")

    @skip_unless_have_test_kmod
    def test_by_pt_regs(self):
        pt_regs = self.prog["drgn_test_kthread_pt_regs"]
        self._test_drgn_test_kthread_trace(self.prog.stack_trace(pt_regs))
        self._test_drgn_test_kthread_trace(self.prog.stack_trace(pt_regs.address_of_()))

    @skip_unless_have_test_kmod
    def test_stack_trace_from_pcs(self):
        if not self.prog["drgn_test_have_stacktrace"]:
            self.skipTest("kernel was not built with CONFIG_STACKTRACE")
        entries = self.prog["drgn_test_stack_entries"]
        self._test_drgn_test_kthread_trace(
            self.prog.stack_trace_from_pcs(
                reinterpret(
                    self.prog.array_type(
                        entries.type_.type,
                        self.prog["drgn_test_num_stack_entries"].value_(),
                    ),
                    entries,
                ).value_()
            )
        )

    @skip_unless_have_test_kmod
    def test_local_variable(self):
        for frame in self.prog.stack_trace(self.prog["drgn_test_kthread"]):
            if frame.name == "drgn_test_kthread_fn3":
                break
        else:
            self.fail("Couldn't find drgn_test_kthread_fn3 frame")
        self.assertEqual(frame["a"], 1)
        self.assertEqual(frame["b"], 2)
        self.assertEqual(frame["c"], 3)

    @skip_unless_have_test_kmod
    def test_locals(self):
        task = self.prog["drgn_test_kthread"]
        stack_trace = self.prog.stack_trace(task)
        for frame in stack_trace:
            if frame.name == "drgn_test_kthread_fn3":
                self.assertSetEqual(set(frame.locals()), {"a", "b", "c", "slab_object"})
                break
        else:
            self.fail("Couldn't find drgn_test_kthread_fn3 frame")

    @unittest.skipUnless(
        NORMALIZED_MACHINE_NAME == "x86_64",
        f"{NORMALIZED_MACHINE_NAME} does not use ORC",
    )
    def test_vmlinux_builtin_orc(self):
        # ORC was introduced in kernel 4.14. Detect the presence of ORC or skip
        # the test.
        try:
            self.prog.symbol("__start_orc_unwind")
        except LookupError:
            ver = self.prog["UTS_RELEASE"].string_().decode()
            self.skipTest(f"ORC is not available for {ver}")

        with drgn_log_level(logging.DEBUG):
            # It is difficult to test stack unwinding in a program without also
            # loading types, which necessarily will also make DWARF CFI and ORC
            # available in the debug file. The way we get around this is by creating
            # a new program with no debuginfo, getting a pt_regs from the program
            # that has debuginfo, and then using that to unwind the kernel. We still
            # need a symbol finder, and we'll need the Module API to recognize the
            # kernel address range correctly.
            prog = Program()
            prog.set_kernel()
            prog.register_symbol_finder(
                "vmlinux_kallsyms", load_vmlinux_kallsyms(prog), enable_index=0
            )
            main = prog.main_module(name="kernel", create=True)
            main.address_range = self.prog.main_module().address_range

            # Luckily, all drgn cares about for x86_64 pt_regs is that it is a
            # structure. Rather than creating a matching struct pt_regs definition,
            # we can just create a dummy one of the correct size:
            #     struct pt_regs { unsigned char[size]; };
            # Drgn will happily use that and reinterpret the bytes correctly.
            real_pt_regs_type = self.prog.type("struct pt_regs")
            fake_pt_regs_type = prog.struct_type(
                tag="pt_regs",
                size=real_pt_regs_type.size,
                members=[
                    TypeMember(
                        prog.array_type(
                            prog.int_type("unsigned char", 1, False),
                            real_pt_regs_type.size,
                        ),
                        "data",
                    ),
                ],
            )

            with fork_and_stop() as pid:
                trace = self.prog.stack_trace(pid)
                regs_dict = trace[0].registers()
                pt_regs_obj = Object(
                    self.prog,
                    real_pt_regs_type,
                    {
                        "bp": regs_dict["rbp"],
                        "sp": regs_dict["rsp"],
                        "ip": regs_dict["rip"],
                        "r15": regs_dict["r15"],
                    },
                )
                fake_pt_regs_obj = Object.from_bytes_(
                    prog, fake_pt_regs_type, pt_regs_obj.to_bytes_()
                )
                # We must set drgn's log level manually, beacuse it won't log messages
                # to the logger if it isn't enabled for them.
                with self.assertLogs("drgn", logging.DEBUG) as log:
                    no_debuginfo_trace = prog.stack_trace(fake_pt_regs_obj)

                dwarf_pcs = []
                for frame in trace:
                    if not dwarf_pcs or dwarf_pcs[-1] != frame.pc:
                        dwarf_pcs.append(frame.pc)
                orc_pcs = [frame.pc for frame in no_debuginfo_trace]
                self.assertEqual(dwarf_pcs, orc_pcs)
                self._check_logged_orc_message(log, "kernel")

    def test_registers(self):
        # Smoke test that we get at least one register and that
        # StackFrame.registers() agrees with StackFrame.register().
        with fork_and_stop() as pid:
            trace = self.prog.stack_trace(pid)
            have_registers = False
            for frame in trace:
                for name, value in frame.registers().items():
                    self.assertEqual(frame.register(name), value)
                    have_registers = True
            self.assertTrue(have_registers)

    def test_sp(self):
        # Smoke test that the stack pointer register shows up in
        # StackFrame.registers().
        with fork_and_stop() as pid:
            trace = self.prog.stack_trace(pid)
            self.assertIn(trace[0].sp, trace[0].registers().values())

    def test_prog(self):
        self.assertEqual(
            self.prog.stack_trace(Object(self.prog, "struct pt_regs", value={})).prog,
            self.prog,
        )

    def test_stack__repr_pretty_(self):
        with fork_and_stop() as pid:
            trace = self.prog.stack_trace(pid)
            assertReprPrettyEqualsStr(trace)
            for frame in trace:
                assertReprPrettyEqualsStr(frame)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import logging

from drgn import Program
from tests import TestCase


# Test that our monkey patch to sync the log level between the logging module
# and libdrgn works.
class TestLogging(TestCase):
    def test_set_level_before(self):
        logger = logging.getLogger("drgn")
        with self.assertLogs(logger, "DEBUG") as cm:
            prog = Program()
            prog._log(0, "foo")
        self.assertIn("DEBUG:drgn:foo", cm.output)

    def test_set_level_after(self):
        prog = Program()
        logger = logging.getLogger("drgn")
        with self.assertLogs(logger, "DEBUG") as cm:
            prog._log(0, "bar")
        self.assertIn("DEBUG:drgn:bar", cm.output)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import re
import sys
from typing import NamedTuple, Sequence, TextIO, Tuple


class ConstantClass(NamedTuple):
    name: str
    enum_class: str
    regex: str
    constants: Sequence[Tuple[str, str]] = ()


CONSTANTS = (
    ConstantClass("AbsenceReason", "Enum", r"DRGN_ABSENCE_REASON_([a-zA-Z0-9_]+)"),
    ConstantClass("Architecture", "Enum", r"DRGN_ARCH_([a-zA-Z0-9_]+)"),
    ConstantClass("FindObjectFlags", "Flag", r"DRGN_FIND_OBJECT_([a-zA-Z0-9_]+)"),
    ConstantClass("KmodSearchMethod", "Enum", r"DRGN_KMOD_SEARCH_([a-zA-Z0-9_]+)"),
    ConstantClass("ModuleFileStatus", "Enum", r"DRGN_MODULE_FILE_([a-zA-Z0-9_]+)"),
    ConstantClass(
        "PlatformFlags",
        "Flag",
        r"DRGN_PLATFORM_([a-zA-Z0-9_]+)(?<!DRGN_PLATFORM_DEFAULT_FLAGS)",
    ),
    ConstantClass("PrimitiveType", "Enum", r"DRGN_(C)_TYPE_([a-zA-Z0-9_]+)"),
    ConstantClass(
        "ProgramFlags", "Flag", r"DRGN_PROGRAM_([a-zA-Z0-9_]+)(?<!DRGN_PROGRAM_ENDIAN)"
    ),
    ConstantClass(
        "Qualifiers", "Flag", r"DRGN_QUALIFIER_([a-zA-Z0-9_]+)", [("NONE", "0")]
    ),
    ConstantClass(
        "SupplementaryFileKind",
        "Enum",
        r"DRGN_SUPPLEMENTARY_FILE_([a-z-A-Z0-9_]+)(?<!DRGN_SUPPLEMENTARY_FILE_NONE)",
    ),
    ConstantClass("SymbolBinding", "Enum", r"DRGN_SYMBOL_BINDING_([a-z-A-Z0-9_]+)"),
    ConstantClass("SymbolKind", "Enum", r"DRGN_SYMBOL_KIND_([a-z-A-Z0-9_]+)"),
    ConstantClass("TypeKind", "Enum", r"DRGN_TYPE_([a-zA-Z0-9_]+)"),
)


def gen_constant_class(
    drgn_h: str, output_file: TextIO, constant_class: ConstantClass
) -> None:
    constants = list(constant_class.constants)
    constants.extend(
        ("_".join(groups[1:]), groups[0])
        for groups in re.findall(
            r"^\s*(" + constant_class.regex + r")\s*[=,]", drgn_h, flags=re.MULTILINE
        )
    )
    output_file.write(
        f"""
static int add_{constant_class.name}(PyObject *m, PyObject *enum_module)
{{
	PyObject *tmp, *item;
	int ret = -1;

	tmp = PyList_New({len(constants)});
	if (!tmp)
		goto out;
"""
    )
    for i, (name, value) in enumerate(constants):
        output_file.write(
            f"""\
	item = Py_BuildValue("sK", "{name}", (unsigned long long){value});
	if (!item)
		goto out;
	PyList_SET_ITEM(tmp, {i}, item);
"""
        )
    output_file.write(
        f"""\
	{constant_class.name}_class = PyObject_CallMethod(enum_module, "{constant_class.enum_class}", "sO", "{constant_class.name}", tmp);
	if (PyModule_AddObjectRef(m, "{constant_class.name}", {constant_class.name}_class))
		goto out;
	Py_DECREF(tmp);
	tmp = PyUnicode_FromString(drgn_{constant_class.name}_DOC);
	if (!tmp)
		goto out;
	if (PyObject_SetAttrString({constant_class.name}_class, "__doc__", tmp) == -1)
		goto out;

	ret = 0;
out:
	Py_XDECREF(tmp);
	return ret;
}}
"""
    )


def gen_constants(input_file: TextIO, output_file: TextIO) -> None:
    drgn_h = input_file.read()
    output_file.write(
        """\
/* Generated by libdrgn/build-aux/gen_constants.py. */

#include "drgnpy.h"

"""
    )
    for constant_class in CONSTANTS:
        output_file.write(f"PyObject *{constant_class.name}_class;\n")
    for constant_class in CONSTANTS:
        gen_constant_class(drgn_h, output_file, constant_class)
    output_file.write(
        """
int add_module_constants(PyObject *m)
{
	PyObject *enum_module;
	int ret;

	enum_module = PyImport_ImportModule("enum");
	if (!enum_module)
		return -1;

"""
    )
    for i, constant_class in enumerate(CONSTANTS):
        if i == 0:
            output_file.write("\tif (")
        else:
            output_file.write("\t    ")
        output_file.write(f"add_{constant_class.name}(m, enum_module) == -1")
        if i == len(CONSTANTS) - 1:
            output_file.write(")\n")
        else:
            output_file.write(" ||\n")
    output_file.write(
        """\
		ret = -1;
	else
		ret = 0;
	Py_DECREF(enum_module);
	return ret;
}
"""
    )


if __name__ == "__main__":
    gen_constants(sys.stdin, sys.stdout)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import argparse
import os
import subprocess
import sys

MANYLINUX_VERSIONS = ("manylinux_2_28", "manylinux2014")


def main() -> None:
    parser = argparse.ArgumentParser(
        description="Build drgn source and manylinux distributions"
    )
    parser.add_argument(
        "--python-version",
        default="",
        help="Python version to build for (e.g., 3.12). "
        "If not given, build for all supported versions",
    )
    parser.add_argument(
        "--manylinux-version",
        dest="manylinux_versions",
        action="append",
        choices=MANYLINUX_VERSIONS,
        help="manylinux version to build for (may be given multiple times). "
        "If not given, build for all supported versions",
    )
    args = parser.parse_args()

    manylinux_versions = args.manylinux_versions or MANYLINUX_VERSIONS

    subprocess.check_call([sys.executable, "setup.py", "sdist"])

    version = subprocess.check_output(
        [sys.executable, "setup.py", "--version"], text=True
    ).strip()
    sdist = f"dist/drgn-{version}.tar.gz"

    docker = os.environ.get("DOCKER")
    if docker is None:
        podman = "podman"
        podman_opts = ["--security-opt", "label=disable"]
    else:
        podman = docker
        uid_gid = f"{os.getuid()}:{os.getgid()}"
        podman_opts = ["--env", f"OWNER={uid_gid}"]

    cwd = os.getcwd()
    for manylinux_version in manylinux_versions:
        plat = manylinux_version + "_x86_64"
        subprocess.check_call(
            [
                podman,
                "run",
                "-it",
                "--env",
                f"PLAT={plat}",
                "--env",
                f"SDIST={sdist}",
                *podman_opts,
                "--volume",
                f"{cwd}:/io:ro",
                "--volume",
                f"{cwd}/dist:/io/dist",
                "--workdir",
                "/io",
                "--hostname",
                "drgn",
                "--rm",
                "--pull",
                "always",
                f"quay.io/pypa/{plat}",
                "./scripts/build_manylinux_in_docker.sh",
                args.python_version,
            ]
        )


if __name__ == "__main__":
    main()



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import errno
import mmap
import os
from pathlib import Path

from drgn.helpers.linux.hugetlb import (
    for_each_hstate,
    huge_page_size,
    hugetlb_total_pages,
    hugetlb_total_usage,
)
from tests.linux_kernel import LinuxKernelTestCase


class TestHugetlb(LinuxKernelTestCase):
    def test_for_each_hstate(self):
        try:
            expected = os.listdir(b"/sys/kernel/mm/hugepages")
        except FileNotFoundError:
            expected = []
        self.assertCountEqual(
            [h.name.string_() for h in for_each_hstate(self.prog)],
            expected,
        )

    def test_huge_page_size(self):
        for hstate in for_each_hstate(self.prog):
            self.assertEqual(
                f"hugepages-{huge_page_size(hstate).value_() // 1024}kB".encode(),
                hstate.name.string_(),
            )
            break
        else:
            self.skipTest("no HugeTLB sizes")

    # Also tests hugetlb_total_pages().
    def test_hugetlb_total_usage(self):
        nr_hugepages_path = Path("/proc/sys/vm/nr_hugepages")
        try:
            old_nr_hugepages = int(nr_hugepages_path.read_text())
        except OSError as e:
            if e.errno not in (errno.ENOENT, errno.EOPNOTSUPP):
                raise
            old_nr_hugepages = None
        try:
            if old_nr_hugepages == 0:
                nr_hugepages_path.write_text("1")

            expected_total = 0
            expected_free = 0
            for path in Path("/sys/kernel/mm/hugepages").glob("hugepages-*kB"):
                pages = (
                    int(path.name[len("hugepages-") : -len("kB")])
                    * 1024
                    // mmap.PAGESIZE
                )
                expected_total += int((path / "nr_hugepages").read_text()) * pages
                expected_free += int((path / "free_hugepages").read_text()) * pages

            usage = hugetlb_total_usage(self.prog)
            self.assertEqual(usage.pages, expected_total)
            self.assertAlmostEqual(usage.free_pages, expected_free, delta=1024)

            self.assertEqual(hugetlb_total_pages(self.prog), expected_total)
        finally:
            if old_nr_hugepages == 0:
                nr_hugepages_path.write_text("0")



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

C_KEYWORDS = (
    "_Atomic",
    "_Bool",
    "char",
    "class",
    "const",
    "double",
    "enum",
    "float",
    "int",
    "long",
    "restrict",
    "short",
    "signed",
    "struct",
    "union",
    "unsigned",
    "void",
    "volatile",
)


def main() -> None:
    keywords = [
        ("C_TOKEN_" + keyword.replace("_", "").upper(), keyword)
        for keyword in C_KEYWORDS
    ]
    print("/* Generated by libdrgn/build-aux/gen_c_keywords_inc_strswitch.py. */")
    print("static const char * const keyword_spelling[] = {")
    for token_kind, keyword in keywords:
        print(f'\t[{token_kind}] = "{keyword}",')
    print("};")
    print()
    print("static int identifier_token_kind(const char *s, size_t len, bool cpp)")
    print("{")
    print("\t@memswitch (s, len)@")
    for token_kind, keyword in keywords:
        print(f'\t@case "{keyword}"@')
        if keyword == "class":
            print(f"\t\treturn cpp ? {token_kind} : C_TOKEN_IDENTIFIER;")
        else:
            print(f"\t\treturn {token_kind};")
    print("\t@default@")
    print("\t\treturn C_TOKEN_IDENTIFIER;")
    print("\t@endswitch@")
    print("}")


if __name__ == "__main__":
    main()



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import re

from drgn import Object
from drgn.helpers.linux.ipc import find_sysv_msg_queue, find_sysv_shm
from tests.linux_kernel.crash_commands import CrashCommandTestCase
from tests.linux_kernel.helpers.test_ipc import tmp_sysv_ipc


class TestIpcs(CrashCommandTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.run_crash_command("set -p")
        cls.ipcs = cls.enterClassContext(tmp_sysv_ipc())

    def test_ipcs(self):
        cmd = self.check_crash_command("ipcs")
        found_shm = {}
        found_sem = {}
        found_msg = {}
        for line in cmd.stdout.splitlines():
            if line.startswith("SHMID_KERNEL"):
                found = found_shm
            elif line.startswith("SEM_ARRAY"):
                found = found_sem
            elif line.startswith("MSG_QUEUE"):
                found = found_msg
            elif match := re.match(r"[0-9a-f]+\s+([0-9a-f]+)\s+([0-9]+)", line):
                found[(int(match.group(1), 16), int(match.group(2)))] = line

        for key, id in self.ipcs.shm:
            self.assertIn((key, id), found_shm)
        self.assertIn("SHM_LOCKED", found_shm[self.ipcs.shm[0]])

        for key, id in self.ipcs.sem:
            self.assertIn((key, id), found_sem)

        for key, id in self.ipcs.msg:
            self.assertIn((key, id), found_msg)

        self.assertEqual(
            cmd.drgn_option.globals["shm"].type_.type_name(), "struct shmid_kernel *"
        )
        self.assertEqual(
            cmd.drgn_option.globals["sem_array"].type_.type_name(), "struct sem_array *"
        )
        self.assertEqual(
            cmd.drgn_option.globals["msg_queue"].type_.type_name(), "struct msg_queue *"
        )

        for variable in (
            "key",
            "id",
            "perms",
            "bytes",
            "nattch",
            "nsems",
            "used_bytes",
            "messages",
        ):
            with self.subTest(variable=variable):
                self.assertIsInstance(cmd.drgn_option.globals[variable], Object)

        self.assertIsInstance(cmd.drgn_option.globals["uid"], int)
        self.assertIsInstance(cmd.drgn_option.globals["status"], str)

    def test_s(self):
        cmd = self.check_crash_command("ipcs -s")
        self.assertIn("SEM_ARRAY", cmd.stdout)
        self.assertNotIn("SHMID_KERNEL", cmd.stdout)
        self.assertNotIn("MSG_QUEUE", cmd.stdout)

        self.assertEqual(
            cmd.drgn_option.globals["sem_array"].type_.type_name(), "struct sem_array *"
        )
        self.assertNotIn("shm", cmd.drgn_option.globals)
        self.assertNotIn("msg_queue", cmd.drgn_option.globals)

        for variable in (
            "key",
            "id",
            "perms",
            "nsems",
        ):
            with self.subTest(variable=variable):
                self.assertIsInstance(cmd.drgn_option.globals[variable], Object)

        self.assertIsInstance(cmd.drgn_option.globals["uid"], int)

        self.assertNotIn("bytes", cmd.drgn_option.globals)
        self.assertNotIn("nattch", cmd.drgn_option.globals)
        self.assertNotIn("used_bytes", cmd.drgn_option.globals)
        self.assertNotIn("messages", cmd.drgn_option.globals)
        self.assertNotIn("status", cmd.drgn_option.globals)

    def test_m(self):
        cmd = self.check_crash_command("ipcs -m")
        self.assertIn("SHMID_KERNEL", cmd.stdout)
        self.assertNotIn("SEM_ARRAY", cmd.stdout)
        self.assertNotIn("MSG_QUEUE", cmd.stdout)

        self.assertEqual(
            cmd.drgn_option.globals["shm"].type_.type_name(), "struct shmid_kernel *"
        )
        self.assertNotIn("sem_array", cmd.drgn_option.globals)
        self.assertNotIn("msg_queue", cmd.drgn_option.globals)

        for variable in (
            "key",
            "id",
            "perms",
            "bytes",
            "nattch",
        ):
            with self.subTest(variable=variable):
                self.assertIsInstance(cmd.drgn_option.globals[variable], Object)

        self.assertIsInstance(cmd.drgn_option.globals["uid"], int)
        self.assertIsInstance(cmd.drgn_option.globals["status"], str)

        self.assertNotIn("nsems", cmd.drgn_option.globals)
        self.assertNotIn("used_bytes", cmd.drgn_option.globals)
        self.assertNotIn("messages", cmd.drgn_option.globals)

    def test_q(self):
        cmd = self.check_crash_command("ipcs -q")
        self.assertIn("MSG_QUEUE", cmd.stdout)
        self.assertNotIn("SHMID_KERNEL", cmd.stdout)
        self.assertNotIn("SEM_ARRAY", cmd.stdout)

        self.assertEqual(
            cmd.drgn_option.globals["msg_queue"].type_.type_name(), "struct msg_queue *"
        )
        self.assertNotIn("shm", cmd.drgn_option.globals)
        self.assertNotIn("sem_array", cmd.drgn_option.globals)

        for variable in (
            "key",
            "id",
            "perms",
            "used_bytes",
            "messages",
        ):
            with self.subTest(variable=variable):
                self.assertIsInstance(cmd.drgn_option.globals[variable], Object)

        self.assertIsInstance(cmd.drgn_option.globals["uid"], int)

        self.assertNotIn("bytes", cmd.drgn_option.globals)
        self.assertNotIn("nattch", cmd.drgn_option.globals)
        self.assertNotIn("status", cmd.drgn_option.globals)
        self.assertNotIn("nsems", cmd.drgn_option.globals)

    def test_id(self):
        shm_key, shm_id = self.ipcs.shm[0]

        cmd = self.check_crash_command(f"ipcs {shm_id}")
        found_shm = {}
        found_sem = {}
        found_msg = {}
        for line in cmd.stdout.splitlines():
            if line.startswith("SHMID_KERNEL"):
                found = found_shm
            elif line.startswith("SEM_ARRAY"):
                found = found_sem
            elif line.startswith("MSG_QUEUE"):
                found = found_msg
            elif match := re.match(r"[0-9a-f]+\s+([0-9a-f]+)\s+([0-9]+)", line):
                found[(int(match.group(1), 16), int(match.group(2)))] = line

        for key, id in self.ipcs.shm:
            if (key, id) == (shm_key, shm_id):
                self.assertIn((key, id), found_shm)
            else:
                self.assertNotIn((key, id), found_shm)
        self.assertIn("SHM_LOCKED", found_shm[self.ipcs.shm[0]])

        self.assertEqual(
            cmd.drgn_option.globals["shm"], find_sysv_shm(self.prog, shm_id)
        )

        for variable in (
            "key",
            "id",
            "perms",
            "bytes",
            "nattch",
        ):
            with self.subTest(variable=variable):
                self.assertIsInstance(cmd.drgn_option.globals[variable], Object)

        self.assertIsInstance(cmd.drgn_option.globals["uid"], int)
        self.assertIsInstance(cmd.drgn_option.globals["status"], str)

    def test_addr(self):
        msg_key, msg_id = self.ipcs.msg[0]
        msg_queue = find_sysv_msg_queue(self.prog, msg_id)

        cmd = self.check_crash_command(f"ipcs {hex(msg_queue)}")
        found_shm = {}
        found_sem = {}
        found_msg = {}
        for line in cmd.stdout.splitlines():
            if line.startswith("SHMID_KERNEL"):
                found = found_shm
            elif line.startswith("SEM_ARRAY"):
                found = found_sem
            elif line.startswith("MSG_QUEUE"):
                found = found_msg
            elif match := re.match(r"[0-9a-f]+\s+([0-9a-f]+)\s+([0-9]+)", line):
                found[(int(match.group(1), 16), int(match.group(2)))] = line

        for key, id in self.ipcs.msg:
            if (key, id) == (msg_key, msg_id):
                self.assertIn((key, id), found_msg)
            else:
                self.assertNotIn((key, id), found_msg)

        self.assertFalse(found_shm)
        self.assertFalse(found_sem)

        self.assertEqual(
            cmd.drgn_option.globals["msg_queue"].type_.type_name(), "struct msg_queue *"
        )

        for variable in (
            "key",
            "id",
            "perms",
            "used_bytes",
            "messages",
        ):
            with self.subTest(variable=variable):
                self.assertIsInstance(cmd.drgn_option.globals[variable], Object)

        self.assertIsInstance(cmd.drgn_option.globals["uid"], int)

        self.assertNotIn("bytes", cmd.drgn_option.globals)
        self.assertNotIn("nattch", cmd.drgn_option.globals)
        self.assertNotIn("status", cmd.drgn_option.globals)
        self.assertNotIn("nsems", cmd.drgn_option.globals)

    def test_n_pid(self):
        cmd = self.check_crash_command("ipcs -n 1")
        self.assertIn("SHM_LOCKED", cmd.stdout)
        self.assertIn("for_each_sysv_shm(ipc_ns)", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["ipc_ns"], self.prog["init_ipc_ns"].address_of_()
        )

    def test_n_task(self):
        task = self.prog["init_task"].address_of_()
        cmd = self.check_crash_command(f"ipcs -n {hex(task)}")
        self.assertIn("SHM_LOCKED", cmd.stdout)
        self.assertIn("for_each_sysv_shm(ipc_ns)", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["ipc_ns"], self.prog["init_ipc_ns"].address_of_()
        )

    def test_n_and_id(self):
        shm_key, shm_id = self.ipcs.shm[0]
        cmd = self.check_crash_command(f"ipcs -n 1 {shm_id}")
        self.assertIn("SHM_LOCKED", cmd.stdout)
        self.assertIn("find_sysv_shm(ipc_ns,", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["ipc_ns"], self.prog["init_ipc_ns"].address_of_()
        )

    def test_n_and_addr(self):
        msg_key, msg_id = self.ipcs.msg[0]
        msg_queue = find_sysv_msg_queue(self.prog, msg_id)

        cmd = self.check_crash_command(f"ipcs -n 1 {hex(msg_queue)}")
        self.assertIn("for_each_sysv_msg_queue(ipc_ns)", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["ipc_ns"], self.prog["init_ipc_ns"].address_of_()
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# Copyright (c) 2026, Oracle and/or its affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later


import os
from pathlib import Path
import re
import signal

from drgn import Object
from drgn.commands import CommandArgumentError
from tests.linux_kernel import fork_and_stop, skip_unless_have_test_kmod
from tests.linux_kernel.crash_commands import CrashCommandTestCase


class TestRunq(CrashCommandTestCase):
    def test_no_args(self):
        cmd = self.check_crash_command("runq")
        self.assertRegex(cmd.stdout, r"CPU \d+ RUNQUEUE [0-9a-f]+")
        self.assertIn("CURRENT:", cmd.stdout)
        self.assertIn("RT PRIO_ARRAY:", cmd.stdout)
        self.assertIn("CFS RB_ROOT:", cmd.stdout)
        self.assertIn("PID:", cmd.stdout)

        self.assertIn("cpu_rq(", cmd.drgn_option.stdout)
        self.assertIn("rq_for_each_rt_task(", cmd.drgn_option.stdout)
        self.assertIn("cfs_rq_for_each_entity(", cmd.drgn_option.stdout)
        self.assertIsInstance(cmd.drgn_option.globals["curr"], Object)

    def test_timestamps(self):
        cmd = self.check_crash_command("runq -t")
        self.assertRegex(cmd.stdout, r"CPU \d+: \d{16}")
        self.assertIn("PID:", cmd.stdout)

        self.assertIn("rq.clock", cmd.drgn_option.stdout)
        self.assertIn("sched_info.last_arrival", cmd.drgn_option.stdout)
        self.assertIsInstance(cmd.drgn_option.globals["rq_clock"], Object)
        self.assertIsInstance(cmd.drgn_option.globals["curr"], Object)
        self.assertIsInstance(cmd.drgn_option.globals["curr_clock"], Object)

    def test_lag(self):
        cmd = self.check_crash_command("runq -T")
        self.assertRegex(cmd.stdout, r"CPU \d+: \d+\.\d+ secs")

        self.assertIsInstance(cmd.drgn_option.globals["rq_clock"], Object)

    def test_elapsed(self):
        cmd = self.check_crash_command("runq -m")
        self.assertRegex(cmd.stdout, r"CPU \d+: \[\s*\d+ \d{2}:\d{2}:\d{2}\.\d{3}\]")
        self.assertIn("PID:", cmd.stdout)

        self.assertIn("rq.clock", cmd.drgn_option.stdout)
        self.assertIn("sched_info.last_arrival", cmd.drgn_option.stdout)
        self.assertIsInstance(cmd.drgn_option.globals["rq_clock"], Object)
        self.assertIsInstance(cmd.drgn_option.globals["curr"], Object)
        self.assertIsInstance(cmd.drgn_option.globals["curr_clock"], Object)

    def test_task_groups(self):
        cmd = self.check_crash_command("runq -g")
        self.assertRegex(cmd.stdout, r"CPU \d+")
        self.assertIn("CURRENT:", cmd.stdout)
        self.assertIn("ROOT_TASK_GROUP:", cmd.stdout)
        self.assertIn("CFS_RQ:", cmd.stdout)
        self.assertIn("PID:", cmd.stdout)

        self.assertIn("cfs_rq_for_each_entity(", cmd.drgn_option.stdout)
        self.assertIn("task_group_name(", cmd.drgn_option.stdout)

    def test_cpu(self):
        cpu = max(os.sched_getaffinity(0))
        cmd = self.check_crash_command(f"runq -c {cpu}")
        self.assertIn(f"CPU {cpu}", cmd.stdout)
        self.assertEqual(cmd.stdout.count("CURRENT:"), 1)


class TestSig(CrashCommandTestCase):
    def _test_sig_drgn_option_common(self, cmd):
        for variable in (
            "pid",
            "command",
            "signal_struct",
            "nr_threads",
            "sigaction",
            "handler",
            "mask",
            "flags",
            "blocked",
            "private_pending",
            "private_pending_signals",
            "shared_pending",
            "shared_pending_signals",
        ):
            with self.subTest(variable=variable):
                self.assertIsInstance(cmd.drgn_option.globals[variable], Object)

        for variable in ("cpu", "signo"):
            with self.subTest(variable=variable):
                self.assertIsInstance(cmd.drgn_option.globals[variable], int)

        self.assertIsInstance(cmd.drgn_option.globals["decoded_flags"], str)

    def test_sig(self):
        with fork_and_stop(
            signal.pthread_sigmask, signal.SIG_BLOCK, {signal.SIGUSR1, signal.SIGUSR2}
        ) as (pid, _):
            os.kill(pid, signal.SIGUSR1)
            os.kill(pid, signal.SIGUSR2)

            sig_status = dict(
                re.findall(
                    r"^(SigBlk|SigPnd|ShdPnd):\s*([0-9a-f]+)",
                    Path(f"/proc/{pid}/status").read_text(),
                    flags=re.M,
                )
            )

            cmd = self.check_crash_command(f"sig {pid}")
            foreach_cmd = self.check_crash_command(f"foreach {pid} sig", mode="capture")

        for c in (cmd, foreach_cmd):
            self.assertIn("PID:", c.stdout)
            self.assertIn("SIGNAL_STRUCT:", c.stdout)
            self.assertIn("NR_THREADS:", c.stdout)

            self.assertEqual(
                re.search(r"^\s*BLOCKED: ([0-9a-f]+)$", c.stdout, flags=re.M).group(1),
                sig_status["SigBlk"],
            )

            for crash_name, proc_name in (
                ("PRIVATE_PENDING", "SigPnd"),
                ("SHARED_PENDING", "ShdPnd"),
            ):
                with self.subTest(set=crash_name):
                    self.assertEqual(
                        re.search(
                            rf"^{crash_name}\n\s*SIGNAL: ([0-9a-f]+)$",
                            c.stdout,
                            flags=re.M,
                        ).group(1),
                        sig_status[proc_name],
                    )
                    sigqueue = re.search(
                        rf"^{crash_name}\n.*\n\s*SIGQUEUE: (.*)", c.stdout, flags=re.M
                    ).group(1)
                    if sig_status[proc_name].replace("0", ""):
                        self.assertRegex(sigqueue, r"^\s*SIG\s+SIGINFO\s*$")
                    else:
                        self.assertEqual(sigqueue, "(empty)")

        self._test_sig_drgn_option_common(cmd)
        for variable in ("sigqueue", "info", "pending_signo"):
            with self.subTest(variable=variable):
                self.assertIsInstance(cmd.drgn_option.globals[variable], Object)

        self.assertEqual(cmd.drgn_option.stdout, foreach_cmd.drgn_option.stdout)

    def test_thread_group(self):
        cmd = self.check_crash_command(f"sig -g {os.getpid()}")
        foreach_cmd = self.check_crash_command(
            f"foreach {os.getpid()} sig -g", mode="capture"
        )

        for c in (cmd, foreach_cmd):
            self.assertEqual(
                len(re.findall(r"^\s*PID:", c.stdout, flags=re.M)),
                len(os.listdir("/proc/self/task")) + 1,
            )

        self.assertIn("for_each_task_in_group(", cmd.drgn_option.stdout)
        self._test_sig_drgn_option_common(cmd)
        # We may not have any pending signals, so don't check for those variables.

        self.assertEqual(cmd.drgn_option.stdout, foreach_cmd.drgn_option.stdout)

    def test_list(self):
        cmd = self.check_crash_command("sig -l")
        self.assertRegex(cmd.stdout, r"(?m)^\s*\[9\] SIGKILL$")

        self.assertIn("signal_numbers(", cmd.drgn_option.stdout)
        self.assertIsInstance(cmd.drgn_option.globals["number"], int)
        self.assertIsInstance(cmd.drgn_option.globals["names"], list)
        self.assertIsInstance(cmd.drgn_option.globals["names"][0], str)

    def test_sigset(self):
        value = (1 << (signal.SIGHUP - 1)) | (1 << (signal.SIGTERM - 1))
        cmd = self.check_crash_command(f"sig -s {value:x}")
        self.assertEqual(cmd.stdout, "SIGHUP SIGTERM\n")
        self.assertIn("decode_sigset(0x", cmd.drgn_option.stdout)
        self.assertEqual(cmd.drgn_option.globals["decoded"], "{SIGHUP,SIGTERM}")


class TestWaitq(CrashCommandTestCase):
    @skip_unless_have_test_kmod
    def test_empty(self):
        cmd = self.check_crash_command("waitq drgn_test_empty_waitq")
        self.assertIn("is empty", cmd.stdout)
        self.assertIn("waitqueue_for_each_task(", cmd.drgn_option.stdout)

    @skip_unless_have_test_kmod
    def test_non_empty(self):
        cmd = self.check_crash_command("waitq drgn_test_waitq")
        self.assertIn('COMMAND: "drgn_test_', cmd.stdout)
        self.assertIn("waitqueue_for_each_task(", cmd.drgn_option.stdout)
        for variable in (
            "task",
            "pid",
            "command",
        ):
            self.assertIsInstance(cmd.drgn_option.globals[variable], Object)
        self.assertIsInstance(cmd.drgn_option.globals["cpu"], int)

    @skip_unless_have_test_kmod
    def test_empty_symbol(self):
        address = self.prog.symbol("drgn_test_empty_waitq").address
        cmd = self.check_crash_command(f"waitq {address:x}")
        self.assertIn("is empty", cmd.stdout)
        self.assertIn("waitqueue_for_each_task(", cmd.drgn_option.stdout)

    @skip_unless_have_test_kmod
    def test_struct(self):
        address = self.prog.symbol("drgn_test_waitq_container").address
        cmd = self.check_crash_command(
            f"waitq drgn_test_waitq_container_struct.waitq {address:x}"
        )
        self.assertIn("is empty", cmd.stdout)
        self.assertIn("waitqueue_for_each_task(", cmd.drgn_option.stdout)

    def test_no_arguments(self):
        self.assertRaisesRegex(
            CommandArgumentError,
            "is required",
            self.run_crash_command,
            "waitq",
        )

    def test_too_many_arguments(self):
        self.assertRaisesRegex(
            CommandArgumentError,
            "unrecognized",
            self.run_crash_command,
            "waitq foo.bar abcd1234 baz",
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import os
import socket

from drgn import cast
from drgn.helpers.linux.fs import fget
from drgn.helpers.linux.pid import find_task
from drgn.helpers.linux.tcp import sk_tcpstate
from tests.linux_kernel import LinuxKernelTestCase, create_socket


class TestTcp(LinuxKernelTestCase):
    def test_sk_tcpstate(self):
        with create_socket() as sock:
            task = find_task(self.prog, os.getpid())
            file = fget(task, sock.fileno())
            sk = cast("struct socket *", file.private_data).sk
            self.assertEqual(sk_tcpstate(sk), self.prog["TCP_CLOSE"])

            sock.bind(("localhost", 0))
            sock.listen()
            self.assertEqual(sk_tcpstate(sk), self.prog["TCP_LISTEN"])

            with socket.create_connection(sock.getsockname()), sock.accept()[
                0
            ] as sock2:
                file = fget(task, sock2.fileno())
                sk = cast("struct socket *", file.private_data).sk
                self.assertEqual(sk_tcpstate(sk), self.prog["TCP_ESTABLISHED"])



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import unittest

from drgn import NULL, Object, sizeof
from drgn.helpers.linux.mapletree import mt_for_each, mtree_load
from drgn.helpers.linux.xarray import xa_is_zero
from tests import slow_test
from tests.linux_kernel import LinuxKernelTestCase, skip_unless_have_test_kmod


@skip_unless_have_test_kmod
class TestMapleTree(LinuxKernelTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        if not cls.prog["drgn_test_have_maple_tree"]:
            raise unittest.SkipTest("kernel does not have maple tree")

    def maple_trees(self, name):
        yield self.prog["drgn_test_maple_tree_" + name].address_of_(), False
        yield self.prog["drgn_test_maple_tree_arange_" + name].address_of_(), True

    def test_mtree_load_empty(self):
        for mt, _ in self.maple_trees("empty"):
            self.assertIdentical(mtree_load(mt, 0), NULL(self.prog, "void *"))
            self.assertIdentical(mtree_load(mt, 666), NULL(self.prog, "void *"))

    def test_mt_for_each_empty(self):
        for mt, _ in self.maple_trees("empty"):
            self.assertIdentical(list(mt_for_each(mt)), [])

    def test_mtree_load_one(self):
        for mt, _ in self.maple_trees("one"):
            self.assertIdentical(mtree_load(mt, 0), NULL(self.prog, "void *"))
            self.assertIdentical(mtree_load(mt, 665), NULL(self.prog, "void *"))
            self.assertIdentical(
                mtree_load(mt, 666), Object(self.prog, "void *", 0xDEADB00)
            )
            self.assertIdentical(mtree_load(mt, 667), NULL(self.prog, "void *"))
            self.assertIdentical(mtree_load(mt, 2**32 - 1), NULL(self.prog, "void *"))
            self.assertIdentical(mtree_load(mt, 2**64 - 1), NULL(self.prog, "void *"))

    def test_mt_for_each_one(self):
        for mt, _ in self.maple_trees("one"):
            self.assertIdentical(
                list(mt_for_each(mt)),
                [(666, 666, Object(self.prog, "void *", 0xDEADB00))],
            )

    def test_mtree_load_one_range(self):
        for mt, _ in self.maple_trees("one_range"):
            self.assertIdentical(mtree_load(mt, 0), NULL(self.prog, "void *"))
            self.assertIdentical(mtree_load(mt, 615), NULL(self.prog, "void *"))
            self.assertIdentical(
                mtree_load(mt, 616), Object(self.prog, "void *", 0xDEADB000)
            )
            self.assertIdentical(
                mtree_load(mt, 660), Object(self.prog, "void *", 0xDEADB000)
            )
            self.assertIdentical(
                mtree_load(mt, 666), Object(self.prog, "void *", 0xDEADB000)
            )
            self.assertIdentical(mtree_load(mt, 667), NULL(self.prog, "void *"))
            self.assertIdentical(mtree_load(mt, 2**32 - 1), NULL(self.prog, "void *"))
            self.assertIdentical(mtree_load(mt, 2**64 - 1), NULL(self.prog, "void *"))

    def test_mt_for_each_one_range(self):
        for mt, _ in self.maple_trees("one_range"):
            self.assertIdentical(
                list(mt_for_each(mt)),
                [(616, 666, Object(self.prog, "void *", 0xDEADB000))],
            )

    def test_mtree_load_one_at_zero(self):
        for mt, _ in self.maple_trees("one_at_zero"):
            self.assertIdentical(mtree_load(mt, 0), Object(self.prog, "void *", 0x1234))
            self.assertIdentical(mtree_load(mt, 1), NULL(self.prog, "void *"))
            self.assertIdentical(mtree_load(mt, 2**32 - 1), NULL(self.prog, "void *"))
            self.assertIdentical(mtree_load(mt, 2**64 - 1), NULL(self.prog, "void *"))

    def test_mt_for_each_one_at_zero(self):
        for mt, _ in self.maple_trees("one_at_zero"):
            self.assertIdentical(
                list(mt_for_each(mt)), [(0, 0, Object(self.prog, "void *", 0x1234))]
            )

    def test_mtree_load_one_range_at_zero(self):
        for mt, _ in self.maple_trees("one_range_at_zero"):
            self.assertIdentical(mtree_load(mt, 0), Object(self.prog, "void *", 0x5678))
            self.assertIdentical(mtree_load(mt, 1), Object(self.prog, "void *", 0x5678))
            self.assertIdentical(
                mtree_load(mt, 0x1336), Object(self.prog, "void *", 0x5678)
            )
            self.assertIdentical(
                mtree_load(mt, 0x1337), Object(self.prog, "void *", 0x5678)
            )
            self.assertIdentical(mtree_load(mt, 0x1338), NULL(self.prog, "void *"))
            self.assertIdentical(mtree_load(mt, 2**32 - 1), NULL(self.prog, "void *"))
            self.assertIdentical(mtree_load(mt, 2**64 - 1), NULL(self.prog, "void *"))

    def test_mt_for_each_one_range_at_zero(self):
        for mt, _ in self.maple_trees("one_range_at_zero"):
            self.assertIdentical(
                list(mt_for_each(mt)),
                [(0, 0x1337, Object(self.prog, "void *", 0x5678))],
            )

    def test_mtree_load_zero_entry(self):
        for mt, _ in self.maple_trees("zero_entry"):
            self.assertIdentical(mtree_load(mt, 666), NULL(self.prog, "void *"))
            self.assertTrue(xa_is_zero(mtree_load(mt, 666, advanced=True)))

    def test_mtree_for_each_zero_entry(self):
        for mt, _ in self.maple_trees("zero_entry"):
            self.assertIdentical(list(mt_for_each(mt)), [])
            entries = list(mt_for_each(mt, advanced=True))
            self.assertEqual(len(entries), 1)
            self.assertEqual(entries[0][:2], (666, 666))
            self.assertTrue(xa_is_zero(entries[0][2]))

    def test_mtree_load_zero_entry_at_zero(self):
        for mt, _ in self.maple_trees("zero_entry_at_zero"):
            self.assertIdentical(mtree_load(mt, 0), NULL(self.prog, "void *"))
            self.assertTrue(xa_is_zero(mtree_load(mt, 0, advanced=True)))

    def test_mtree_for_each_zero_entry_at_zero(self):
        for mt, _ in self.maple_trees("zero_entry_at_zero"):
            self.assertIdentical(list(mt_for_each(mt)), [])
            entries = list(mt_for_each(mt, advanced=True))
            self.assertEqual(len(entries), 1)
            self.assertEqual(entries[0][:2], (0, 0))
            self.assertTrue(xa_is_zero(entries[0][2]))

    def test_mtree_load_dense(self):
        for mt, _ in self.maple_trees("dense"):
            for i in range(5):
                with self.subTest(i=i):
                    self.assertIdentical(
                        mtree_load(mt, i), Object(self.prog, "void *", 0xB0BA000 | i)
                    )
            self.assertIdentical(mtree_load(mt, 5), NULL(self.prog, "void *"))
            self.assertIdentical(mtree_load(mt, 2**32 - 1), NULL(self.prog, "void *"))
            self.assertIdentical(mtree_load(mt, 2**64 - 1), NULL(self.prog, "void *"))

    def test_mt_for_each_dense(self):
        for mt, _ in self.maple_trees("dense"):
            self.assertIdentical(
                list(mt_for_each(mt)),
                [(i, i, Object(self.prog, "void *", 0xB0BA000 | i)) for i in range(5)],
            )

    def test_mtree_load_dense_ranges(self):
        for mt, _ in self.maple_trees("dense_ranges"):
            self.assertIdentical(
                mtree_load(mt, 0), Object(self.prog, "void *", 0xB0BA000)
            )
            for i in range(1, 4):
                self.assertIdentical(
                    mtree_load(mt, i), Object(self.prog, "void *", 0xB0BA001)
                )
            for i in range(4, 9):
                self.assertIdentical(
                    mtree_load(mt, i), Object(self.prog, "void *", 0xB0BA002)
                )
            for i in range(9, 16):
                self.assertIdentical(
                    mtree_load(mt, i), Object(self.prog, "void *", 0xB0BA003)
                )
            for i in range(16, 25):
                self.assertIdentical(
                    mtree_load(mt, i), Object(self.prog, "void *", 0xB0BA004)
                )
            self.assertIdentical(mtree_load(mt, 25), NULL(self.prog, "void *"))
            self.assertIdentical(mtree_load(mt, 2**32 - 1), NULL(self.prog, "void *"))
            self.assertIdentical(mtree_load(mt, 2**64 - 1), NULL(self.prog, "void *"))

    def test_mt_for_each_dense_ranges(self):
        for mt, _ in self.maple_trees("dense_ranges"):
            self.assertIdentical(
                list(mt_for_each(mt)),
                [
                    (
                        i**2,
                        (i + 1) ** 2 - 1,
                        Object(self.prog, "void *", 0xB0BA000 | i),
                    )
                    for i in range(5)
                ],
            )

    def test_mtree_load_sparse(self):
        for mt, _ in self.maple_trees("sparse"):
            self.assertIdentical(mtree_load(mt, 0), NULL(self.prog, "void *"))
            self.assertIdentical(
                mtree_load(mt, 1), Object(self.prog, "void *", 0xB0BA000)
            )
            for i in range(2, 4):
                self.assertIdentical(mtree_load(mt, i), NULL(self.prog, "void *"))
            self.assertIdentical(
                mtree_load(mt, 4), Object(self.prog, "void *", 0xB0BA001)
            )
            for i in range(5, 9):
                self.assertIdentical(mtree_load(mt, i), NULL(self.prog, "void *"))
            self.assertIdentical(mtree_load(mt, 26), NULL(self.prog, "void *"))
            self.assertIdentical(mtree_load(mt, 2**32 - 1), NULL(self.prog, "void *"))
            self.assertIdentical(mtree_load(mt, 2**64 - 1), NULL(self.prog, "void *"))

    def test_mt_for_each_sparse(self):
        for mt, _ in self.maple_trees("sparse"):
            self.assertIdentical(
                list(mt_for_each(mt)),
                [
                    (
                        (i + 1) ** 2,
                        (i + 1) ** 2,
                        Object(self.prog, "void *", 0xB0BA000 | i),
                    )
                    for i in range(5)
                ],
            )

    def test_mtree_load_sparse_ranges(self):
        for mt, _ in self.maple_trees("sparse_ranges"):
            self.assertIdentical(mtree_load(mt, 0), NULL(self.prog, "void *"))
            for i in range(1, 5):
                self.assertIdentical(
                    mtree_load(mt, i), Object(self.prog, "void *", 0xB0BA000)
                )
            for i in range(5, 9):
                self.assertIdentical(mtree_load(mt, i), NULL(self.prog, "void *"))
            for i in range(9, 17):
                self.assertIdentical(
                    mtree_load(mt, i), Object(self.prog, "void *", 0xB0BA001)
                )
            for i in range(17, 25):
                self.assertIdentical(mtree_load(mt, i), NULL(self.prog, "void *"))
            for i in range(25, 37):
                self.assertIdentical(
                    mtree_load(mt, i), Object(self.prog, "void *", 0xB0BA002)
                )
            for i in range(37, 49):
                self.assertIdentical(mtree_load(mt, i), NULL(self.prog, "void *"))
            for i in range(49, 65):
                self.assertIdentical(
                    mtree_load(mt, i), Object(self.prog, "void *", 0xB0BA003)
                )
            for i in range(65, 81):
                self.assertIdentical(mtree_load(mt, i), NULL(self.prog, "void *"))
            for i in range(81, 101):
                self.assertIdentical(
                    mtree_load(mt, i), Object(self.prog, "void *", 0xB0BA004)
                )
            self.assertIdentical(mtree_load(mt, 101), NULL(self.prog, "void *"))
            self.assertIdentical(mtree_load(mt, 2**32 - 1), NULL(self.prog, "void *"))
            self.assertIdentical(mtree_load(mt, 2**64 - 1), NULL(self.prog, "void *"))

    def test_mt_for_each_sparse_ranges(self):
        for mt, _ in self.maple_trees("sparse_ranges"):
            self.assertIdentical(
                list(mt_for_each(mt)),
                [
                    (
                        (2 * i + 1) ** 2,
                        (2 * i + 2) ** 2,
                        Object(self.prog, "void *", 0xB0BA000 | i),
                    )
                    for i in range(5)
                ],
            )

    @slow_test
    def test_mtree_load_three_levels_dense_1(self):
        maple_range64_slots = self.prog["drgn_test_maple_range64_slots"].value_()
        ulong_max = (1 << (sizeof(self.prog.type("unsigned long")) * 8)) - 1
        for mt, arange in self.maple_trees("three_levels_dense_1"):
            node_slots = self.prog[
                (
                    "drgn_test_maple_arange64_slots"
                    if arange
                    else "drgn_test_maple_range64_slots"
                )
            ].value_()
            n = 2 * (node_slots - 1) * (maple_range64_slots - 1) + (
                maple_range64_slots - 1
            )
            for i in range(n):
                with self.subTest(i=i):
                    self.assertIdentical(
                        mtree_load(mt, i), Object(self.prog, "void *", 0xB0BA000 | i)
                    )
            self.assertIdentical(mtree_load(mt, n), NULL(self.prog, "void *"))
            self.assertIdentical(mtree_load(mt, n + 1), NULL(self.prog, "void *"))
            self.assertIdentical(
                mtree_load(mt, ulong_max - 1), NULL(self.prog, "void *")
            )
            self.assertIdentical(mtree_load(mt, ulong_max), NULL(self.prog, "void *"))

    def test_mt_for_each_three_levels_dense_1(self):
        maple_range64_slots = self.prog["drgn_test_maple_range64_slots"].value_()
        for mt, arange in self.maple_trees("three_levels_dense_1"):
            node_slots = self.prog[
                (
                    "drgn_test_maple_arange64_slots"
                    if arange
                    else "drgn_test_maple_range64_slots"
                )
            ].value_()
            n = 2 * (node_slots - 1) * (maple_range64_slots - 1) + (
                maple_range64_slots - 1
            )
            self.assertIdentical(
                list(mt_for_each(mt)),
                [(i, i, Object(self.prog, "void *", 0xB0BA000 | i)) for i in range(n)],
            )

    @slow_test
    def test_mtree_load_three_levels_dense_2(self):
        maple_range64_slots = self.prog["drgn_test_maple_range64_slots"].value_()
        ulong_max = (1 << (sizeof(self.prog.type("unsigned long")) * 8)) - 1
        for mt, arange in self.maple_trees("three_levels_dense_2"):
            node_slots = self.prog[
                (
                    "drgn_test_maple_arange64_slots"
                    if arange
                    else "drgn_test_maple_range64_slots"
                )
            ].value_()
            n = 2 * node_slots * maple_range64_slots
            for i in range(n):
                with self.subTest(i=i):
                    self.assertIdentical(
                        mtree_load(mt, i), Object(self.prog, "void *", 0xB0BA000 | i)
                    )
            self.assertIdentical(mtree_load(mt, n), NULL(self.prog, "void *"))
            self.assertIdentical(mtree_load(mt, n + 1), NULL(self.prog, "void *"))
            self.assertIdentical(
                mtree_load(mt, ulong_max - 1), NULL(self.prog, "void *")
            )
            self.assertIdentical(mtree_load(mt, ulong_max), NULL(self.prog, "void *"))

    def test_mt_for_each_three_levels_dense_2(self):
        maple_range64_slots = self.prog["drgn_test_maple_range64_slots"].value_()
        for mt, arange in self.maple_trees("three_levels_dense_2"):
            node_slots = self.prog[
                (
                    "drgn_test_maple_arange64_slots"
                    if arange
                    else "drgn_test_maple_range64_slots"
                )
            ].value_()
            n = 2 * node_slots * maple_range64_slots
            self.assertIdentical(
                list(mt_for_each(mt)),
                [(i, i, Object(self.prog, "void *", 0xB0BA000 | i)) for i in range(n)],
            )

    @slow_test
    def test_mtree_load_three_levels_ranges_1(self):
        maple_range64_slots = self.prog["drgn_test_maple_range64_slots"].value_()
        ulong_max = (1 << (sizeof(self.prog.type("unsigned long")) * 8)) - 1
        for mt, arange in self.maple_trees("three_levels_ranges_1"):
            node_slots = self.prog[
                (
                    "drgn_test_maple_arange64_slots"
                    if arange
                    else "drgn_test_maple_range64_slots"
                )
            ].value_()
            n = 2 * (node_slots - 1) * (maple_range64_slots - 1) + (
                maple_range64_slots - 1
            )
            for i in range(n):
                with self.subTest(i=i):
                    self.assertIdentical(
                        mtree_load(mt, 2 * i),
                        Object(self.prog, "void *", 0xB0BA000 | i),
                    )
                    self.assertIdentical(
                        mtree_load(mt, 2 * i + 1),
                        Object(self.prog, "void *", 0xB0BA000 | i),
                    )
            self.assertIdentical(
                mtree_load(mt, 2 * n), Object(self.prog, "void *", 0xB0BA000 | n)
            )
            self.assertIdentical(
                mtree_load(mt, 2 * n + 1), Object(self.prog, "void *", 0xB0BA000 | n)
            )
            self.assertIdentical(
                mtree_load(mt, ulong_max - 1),
                Object(self.prog, "void *", 0xB0BA000 | n),
            )
            self.assertIdentical(
                mtree_load(mt, ulong_max), Object(self.prog, "void *", 0xB0BA000 | n)
            )

    def test_mt_for_each_three_levels_ranges_1(self):
        maple_range64_slots = self.prog["drgn_test_maple_range64_slots"].value_()
        ulong_max = (1 << (sizeof(self.prog.type("unsigned long")) * 8)) - 1
        for mt, arange in self.maple_trees("three_levels_ranges_1"):
            node_slots = self.prog[
                (
                    "drgn_test_maple_arange64_slots"
                    if arange
                    else "drgn_test_maple_range64_slots"
                )
            ].value_()
            n = 2 * (node_slots - 1) * (maple_range64_slots - 1) + (
                maple_range64_slots - 1
            )
            self.assertIdentical(
                list(mt_for_each(mt)),
                [
                    (2 * i, 2 * i + 1, Object(self.prog, "void *", 0xB0BA000 | i))
                    for i in range(n)
                ]
                + [(2 * n, ulong_max, Object(self.prog, "void *", 0xB0BA000 | n))],
            )

    @slow_test
    def test_mtree_load_three_levels_ranges_2(self):
        maple_range64_slots = self.prog["drgn_test_maple_range64_slots"].value_()
        ulong_max = (1 << (sizeof(self.prog.type("unsigned long")) * 8)) - 1
        for mt, arange in self.maple_trees("three_levels_ranges_2"):
            node_slots = self.prog[
                (
                    "drgn_test_maple_arange64_slots"
                    if arange
                    else "drgn_test_maple_range64_slots"
                )
            ].value_()
            n = 2 * node_slots * maple_range64_slots
            for i in range(n):
                with self.subTest(i=i):
                    self.assertIdentical(
                        mtree_load(mt, 2 * i),
                        Object(self.prog, "void *", 0xB0BA000 | i),
                    )
                    self.assertIdentical(
                        mtree_load(mt, 2 * i + 1),
                        Object(self.prog, "void *", 0xB0BA000 | i),
                    )
            self.assertIdentical(
                mtree_load(mt, 2 * n), Object(self.prog, "void *", 0xB0BA000 | n)
            )
            self.assertIdentical(
                mtree_load(mt, 2 * n + 1), Object(self.prog, "void *", 0xB0BA000 | n)
            )
            self.assertIdentical(
                mtree_load(mt, ulong_max - 1),
                Object(self.prog, "void *", 0xB0BA000 | n),
            )
            self.assertIdentical(
                mtree_load(mt, ulong_max), Object(self.prog, "void *", 0xB0BA000 | n)
            )

    def test_mt_for_each_three_levels_ranges_2(self):
        maple_range64_slots = self.prog["drgn_test_maple_range64_slots"].value_()
        ulong_max = (1 << (sizeof(self.prog.type("unsigned long")) * 8)) - 1
        for mt, arange in self.maple_trees("three_levels_ranges_2"):
            node_slots = self.prog[
                (
                    "drgn_test_maple_arange64_slots"
                    if arange
                    else "drgn_test_maple_range64_slots"
                )
            ].value_()
            n = 2 * node_slots * maple_range64_slots
            self.assertIdentical(
                list(mt_for_each(mt)),
                [
                    (2 * i, 2 * i + 1, Object(self.prog, "void *", 0xB0BA000 | i))
                    for i in range(n)
                ]
                + [(2 * n, ulong_max, Object(self.prog, "void *", 0xB0BA000 | n))],
            )
